from .exceptions import ChangeRadiiError, ParmWarning
import warnings

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def bondi(parm):
    """ Sets the bondi radii """
    for i, atom in enumerate(parm.atoms):
        # Radius of C atom depends on what type it is
        if atom.atomic_number == 6:
            if isinstance(atom.type, int):
                atom.solvent_radius = 1.7
            elif atom.type.startswith('C1') and atom.mass > 13.0:
                atom.solvent_radius = 2.2
            elif atom.type.startswith('C2') and atom.mass > 14.0:
                atom.solvent_radius = 2.2
            elif atom.type.startswith('C3') and atom.mass > 15.0:
                atom.solvent_radius = 2.2
            else:
                atom.solvent_radius = 1.7
        # All other elements have fixed radii for all types/partners
        elif atom.atomic_number == 1:
            atom.solvent_radius = 1.2
        elif atom.atomic_number == 7:
            atom.solvent_radius = 1.55
        elif atom.atomic_number == 8:
            atom.solvent_radius = 1.5
        elif atom.atomic_number == 9:
            atom.solvent_radius = 1.5
        elif atom.atomic_number == 14:
            atom.solvent_radius = 2.1
        elif atom.atomic_number == 15:
            atom.solvent_radius = 1.85
        elif atom.atomic_number == 16:
            atom.solvent_radius = 1.8
        elif atom.atomic_number == 17:
            atom.solvent_radius = 1.7
        else:
            atom.solvent_radius = 1.5

    try:
        parm.parm_data['RADIUS_SET'][0] = 'Bondi radii (bondi)'
    except AttributeError:
        pass
    _screen1(parm)

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def amber6(parm):
    """ Sets the amber6 radii """
    for i, atom in enumerate(parm.atoms):
        # Radius of H atom depends on element it is bonded to
        bondeds = list(atom.bond_partners)
        if atom.atomic_number == 1:
            if not bondeds:
                warnings.warn(f"hydrogen atom {i} has no bonds; this is unusual", ParmWarning)
                atom.solvent_radius = 1.2
            elif bondeds[0].atomic_number == 6: # carbon
                atom.solvent_radius = 1.3
            elif bondeds[0].atomic_number in (8, 16): # oxygen or sulfur
                atom.solvent_radius = 0.8
            else: # anything else
                atom.solvent_radius = 1.2
        # Radius of C atom depends on what type it is
        elif atom.atomic_number == 6:
            if isinstance(atom.type, int):
                atom.solvent_radius = 1.7
            elif atom.type.startswith('C1') and atom.mass > 13.0:
                atom.solvent_radius = 2.2
            elif atom.type.startswith('C2') and atom.mass > 14.0:
                atom.solvent_radius = 2.2
            elif atom.type.startswith('C3') and atom.mass > 15.0:
                atom.solvent_radius = 2.2
            else:
                atom.solvent_radius = 1.7
        # All other elements have fixed solvent_radius for all types/partners
        elif atom.atomic_number == 7:
            atom.solvent_radius = 1.55
        elif atom.atomic_number == 8:
            atom.solvent_radius = 1.5
        elif atom.atomic_number == 9:
            atom.solvent_radius = 1.5
        elif atom.atomic_number == 14:
            atom.solvent_radius = 2.1
        elif atom.atomic_number == 15:
            atom.solvent_radius = 1.85
        elif atom.atomic_number == 16:
            atom.solvent_radius = 1.8
        elif atom.atomic_number == 17:
            atom.solvent_radius = 1.7
        else:
            atom.solvent_radius = 1.5
    try:
        parm.parm_data['RADIUS_SET'][0] = 'amber6 modified Bondi radii (amber6)'
    except AttributeError:
        pass
    _screen1(parm)

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def mbondi(parm):
    """ Sets the mbondi radii """
    for i, atom in enumerate(parm.atoms):
        # Radius of H atom depends on element it is bonded to
        if atom.atomic_number == 1:
            bondeds = list(atom.bond_partners)
            if not bondeds:
                warnings.warn(f"hydrogen atom {i} has no bonds; this is unusual", ParmWarning)
                atom.solvent_radius = 0.8
            elif bondeds[0].atomic_number in (6, 7): # C or N
                atom.solvent_radius = 1.3
            elif bondeds[0].atomic_number in (8, 16): # O or S
                atom.solvent_radius = 0.8
            else:
                atom.solvent_radius = 1.2
        # Radius of C atom depends on what type it is
        elif atom.atomic_number == 6:
            if isinstance(atom.type, int):
                atom.solvent_radius = 1.7
            elif atom.type.startswith('C1') and atom.mass > 13.0:
                atom.solvent_radius = 2.2
            elif atom.type.startswith('C2') and atom.mass > 14.0:
                atom.solvent_radius = 2.2
            elif atom.type.startswith('C3') and atom.mass > 15.0:
                atom.solvent_radius = 2.2
            else:
                atom.solvent_radius = 1.7
        # All other elements have fixed solvent_radius for all types/partners
        elif atom.atomic_number == 7:
            atom.solvent_radius = 1.55
        elif atom.atomic_number == 8:
            atom.solvent_radius = 1.5
        elif atom.atomic_number == 9:
            atom.solvent_radius = 1.5
        elif atom.atomic_number == 14:
            atom.solvent_radius = 2.1
        elif atom.atomic_number == 15:
            atom.solvent_radius = 1.85
        elif atom.atomic_number == 16:
            atom.solvent_radius = 1.8
        elif atom.atomic_number == 17:
            atom.solvent_radius = 1.7
        else:
            atom.solvent_radius = 1.5
    try:
        parm.parm_data['RADIUS_SET'][0] = 'modified Bondi radii (mbondi)'
    except AttributeError:
        pass
    _screen1(parm)

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def mbondi_pb2(parm):
    """
    Sets the mbondi_pb2 radii

    Use ropt values for halogens in Table 3 (https://pubs.acs.org/doi/full/10.1021/acs.jctc.9b00106)

                     pb2
    halogen	 rstd	 MAEstd	 MAEopt (ropt)
    Cl	     1.70    1.544	 1.529  (1.76)
    Br	     1.85    0.792	 0.743  (1.97)
    I	     1.98    0.905	 0.858  (2.09)

    These ropt values have been obtained using the "pb2" routine described
    in https://pubs.acs.org/doi/full/10.1021/acs.jctc.9b00106

    setup	radii set(s)	ΔGnonpolar	      γ	        β        inp value in gmx_MMPBSA
    pb2	    mbondi	        γ * SASA + β	  0.00720	0.0000   1

    """
    for i, atom in enumerate(parm.atoms):
        # Radius of H atom depends on element it is bonded to
        if atom.atomic_number == 1:
            bondeds = list(atom.bond_partners)
            if bondeds[0].atomic_number in (6, 7): # C or N
                atom.solvent_radius = 1.3
            elif bondeds[0].atomic_number in (8, 16): # O or S
                atom.solvent_radius = 0.8
            else:
                atom.solvent_radius = 1.2
        # Radius of C atom depends on what type it is
        elif atom.atomic_number == 6:
            if atom.type.startswith('C1') and atom.mass > 13.0:
                atom.solvent_radius = 2.2
            if atom.type.startswith('C2') and atom.mass > 14.0:
                atom.solvent_radius = 2.2
            if atom.type.startswith('C3') and atom.mass > 15.0:
                atom.solvent_radius = 2.2
            else:
                atom.solvent_radius = 1.7
        # All other elements have fixed solvent_radius for all types/partners
        elif atom.atomic_number == 7:
            atom.solvent_radius = 1.55
        elif atom.atomic_number == 8:
            atom.solvent_radius = 1.5
        elif atom.atomic_number == 9:
            atom.solvent_radius = 1.5
        elif atom.atomic_number == 14:
            atom.solvent_radius = 2.1
        elif atom.atomic_number == 15:
            atom.solvent_radius = 1.85
        elif atom.atomic_number == 16:
            atom.solvent_radius = 1.8
        # Cl
        elif atom.atomic_number == 17:
            atom.solvent_radius = 1.76
        # Br
        elif atom.atomic_number == 35:
            atom.solvent_radius = 1.97
        # I
        elif atom.atomic_number == 53:
            atom.solvent_radius = 2.09
        else:
            atom.solvent_radius = 1.5
    try:
        parm.parm_data['RADIUS_SET'][0] = 'modified Bondi radii with ropt for halogens (mbondi_pb2)'
    except AttributeError:
        pass
    _screen1(parm)


def mbondi_pb3(parm):
    """
    Sets the mbondi_pb3 radii

    Use ropt values for halogens in Table 3 (https://pubs.acs.org/doi/full/10.1021/acs.jctc.9b00106)

                     pb3
    halogen	 rstd	 MAEstd	 MAEopt (ropt)
    Cl	     1.70    1.358	 1.236 (2.20)
    Br	     1.85    0.770	 0.681 (2.04)
    I	     1.98    0.769	 0.697 (2.19)

    These ropt values have been obtained using the "pb3" routine described
    in https://pubs.acs.org/doi/full/10.1021/acs.jctc.9b00106

    setup	radii set(s)	ΔGnonpolar	                γ	     β        inp value in gmx_MMPBSA
    pb3	    mbondi + Rmin	γ * SAV + β + ΔGdispersion	0.03780	 –0.5692  2

    The mbondi radii set is used in the ΔGpolar calculation, while the value corresponding to Rmin in GAFF is
    used for the calculation of ΔGnonpolar

    """
    for i, atom in enumerate(parm.atoms):
        # Radius of H atom depends on element it is bonded to
        if atom.atomic_number == 1:
            bondeds = list(atom.bond_partners)
            if bondeds[0].atomic_number in (6, 7): # C or N
                atom.solvent_radius = 1.3
            elif bondeds[0].atomic_number in (8, 16): # O or S
                atom.solvent_radius = 0.8
            else:
                atom.solvent_radius = 1.2
        # Radius of C atom depends on what type it is
        elif atom.atomic_number == 6:
            if atom.type.startswith('C1') and atom.mass > 13.0:
                atom.solvent_radius = 2.2
            if atom.type.startswith('C2') and atom.mass > 14.0:
                atom.solvent_radius = 2.2
            if atom.type.startswith('C3') and atom.mass > 15.0:
                atom.solvent_radius = 2.2
            else:
                atom.solvent_radius = 1.7
        # All other elements have fixed solvent_radius for all types/partners
        elif atom.atomic_number == 7:
            atom.solvent_radius = 1.55
        elif atom.atomic_number == 8:
            atom.solvent_radius = 1.5
        elif atom.atomic_number == 9:
            atom.solvent_radius = 1.5
        elif atom.atomic_number == 14:
            atom.solvent_radius = 2.1
        elif atom.atomic_number == 15:
            atom.solvent_radius = 1.85
        elif atom.atomic_number == 16:
            atom.solvent_radius = 1.8
        # Cl
        elif atom.atomic_number == 17:
            atom.solvent_radius = 2.20
        # Br
        elif atom.atomic_number == 35:
            atom.solvent_radius = 2.04
        # I
        elif atom.atomic_number == 53:
            atom.solvent_radius = 2.19
        else:
            atom.solvent_radius = 1.5
    try:
        parm.parm_data['RADIUS_SET'][0] = 'modified Bondi radii with ropt for halogens (mbondi_pb3)'
    except AttributeError:
        pass
    _screen1(parm)

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def mbondi2(parm):
    """ Sets the mbondi2 radii """
    for i, atom in enumerate(parm.atoms):
        # Radius of H atom depends on element it is bonded to
        if atom.atomic_number == 1:
            if not atom.bond_partners:
                warnings.warn(f"hydrogen atom {i} has no bonds; this is unusual", ParmWarning)
                atom.solvent_radius = 1.2
            if atom.bond_partners[0].atomic_number == 7:
                atom.solvent_radius = 1.3
            else:
                atom.solvent_radius = 1.2
        # Radius of C atom depends on what type it is
        elif atom.atomic_number == 6:
            if isinstance(atom.type, int):
                atom.solvent_radius = 1.7
            elif atom.type.startswith('C1') and atom.mass > 13.0:
                atom.solvent_radius = 2.2
            elif atom.type.startswith('C2') and atom.mass > 14.0:
                atom.solvent_radius = 2.2
            elif atom.type.startswith('C3') and atom.mass > 15.0:
                atom.solvent_radius = 2.2
            else:
                atom.solvent_radius = 1.7
        # All other elements have fixed solvent_radius for all types/partners
        elif atom.atomic_number == 7:
            atom.solvent_radius = 1.55
        elif atom.atomic_number == 8:
            atom.solvent_radius = 1.5
        elif atom.atomic_number == 9:
            atom.solvent_radius = 1.5
        elif atom.atomic_number == 14:
            atom.solvent_radius = 2.1
        elif atom.atomic_number == 15:
            atom.solvent_radius = 1.85
        elif atom.atomic_number == 16:
            atom.solvent_radius = 1.8
        elif atom.atomic_number == 17:
            atom.solvent_radius = 1.7
        else:
            atom.solvent_radius = 1.5
    try:
        parm.parm_data['RADIUS_SET'][0] = 'H(N)-modified Bondi radii (mbondi2)'
    except AttributeError:
        pass
    _screen1(parm)

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def mbondi3(parm):
    """ Sets mbondi3 radii """
    mbondi2(parm) # start from mbondi2 radii
    for i, atom in enumerate(parm.atoms):
        # Adjust OE (GLU), OD (ASP), and HH/HE (ARG)
        if atom.residue.name in ('GLU', 'ASP', 'GL4', 'AS4'):
            if atom.name.startswith('OE') or atom.name.startswith('OD'):
                atom.solvent_radius = 1.4
        elif atom.residue.name == 'ARG':
            if atom.name.startswith('HH') or atom.name.startswith('HE'):
                atom.solvent_radius = 1.17
        # Adjust carboxylate O radii on C-Termini. Don't just do the end
        # residue, since we can have C-termini in the middle as well
        # (i.e., 2-chain dimers)
        if atom.name == 'OXT':
            atom.solvent_radius = 1.4
            parm.atoms[i-1].solvent_radius = 1.4

    try:
        parm.parm_data['RADIUS_SET'][0] = 'ArgH and AspGluO modified Bondi2 radii (mbondi3)'
    except AttributeError:
        pass
    _screen1(parm)

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def charmm_radii(parm):
    """
    Sets charmm radii

    * Atomic radii for Poisson-Boltzmann calculations derived from average
    * solvent electrostatic charge distribution with explicit solvent.
    * Accuracy tested with free energy perturbation with explicit solvent

    * Most of the values were taken from a *_radii.str file used in PBEQ Solver in charmm-gui

    ! Radii for protein atoms in 20 standard amino acids
    ! Authors:  Mafalda Nina, Dmitrii Belogv, and Benoit Roux
    ! University of Montreal, June 1996.
    ! M. Nina and B. Roux. "Atomic Radii for Continuum Electrostatics
    ! Calculations Based on Molecular Dynamics Free Energy Simulations",
    ! J. Phys. Chem. B 101: 5239-5248 (1997).

    ! Radii for nucleic acid atoms (RNA and DNA)
    ! Authors:  Nilesh Banavali and Benoit Roux
    ! Cornell University, June 2002.
    ! N.K. Banavali and B. Roux, "Atomic Radii for Continuum Electrostatics
    ! Calculations on Nucleic Acids", J. Phys.  Chem. B 106:11026-11035 (2002).

    ! UPDATES:
    ! --------
    !
    ! December 1998:  GLU and ASP modified by Mafalda Nina.
    ! January  1999:  Protonated histidine HSP has been included by Mafalda Nina,
    !                 dG_elec = -68.15 kcal/mol (MD/FES) or -68.10 kcal/mol (PBEQ).
    ! January  1999:  TEA and ions added by Benoit Roux.
    ! November 2000:  Na+ added by Benoit Roux.
    ! June     2001:  All nucleic acid atoms added by Nilesh Banavali.
    ! February 2022:  Other atoms (https://pubs.acs.org/doi/10.1021/acs.jcim.1c00177)

    """
    for i, atom in enumerate(parm.atoms):
        # PROTEIN ATOMS
        if atom.residue.name in ['ALA', 'ARG', 'ASN', 'ASP', 'ASPP', 'CYS', 'GLN', 'GLU', 'GLUP', 'GLY', 'HSD',
                                 'HSE', 'HSP', 'ILE', 'LEU', 'LYS', 'LSN', 'MET', 'PHE', 'PRO', 'SER', 'SERD',
                                 'THR', 'TRP', 'TYR', 'VAL', 'ALAD', 'CYM', 'CYSD']:
            # !Patches CT3 N-Methylamide C-terminus
            # !        ACE acetylated N-terminus (ACP for PRO)
            # scalar wmain set 2.06 sele (type CAY .or. type CAT) end
            if atom.name == 'CAY' or atom.name == 'CAT':
                atom.solvent_radius = 2.06
            # scalar wmain set 2.04 sele type CY end
            if atom.name == 'CY':
                atom.solvent_radius = 2.04
            # scalar wmain set 1.52 sele type OY end
            if atom.name == 'OY':
                atom.solvent_radius = 1.52
            # scalar wmain set 2.23 sele type NT end
            if atom.name == 'NT':
                atom.solvent_radius = 2.23
            # ! for COO- terminus
            # scalar wmain set 1.40 sele type OT* end
            if atom.name.startswith('OT'):
                atom.solvent_radius = 1.40

            # !Backbone
            # scalar wmain set 2.04 sele type C  end
            #        ! for peptide bond
            if atom.name == 'C':
                atom.solvent_radius = 2.04
            # scalar wmain set 1.52 sele type O  end
            #        ! for peptide bond
            if atom.name == 'O':
                atom.solvent_radius = 1.52
            # scalar wmain set 2.23 sele type N  end
            #        ! for peptide bond
            if atom.name == 'N':
                atom.solvent_radius = 2.23
            # scalar wmain set 2.86 sele type CA  end
            #        ! for all CA except GLY
            # scalar wmain set 2.38 sele (resnam GLY .and. type CA) end
            # scalar wmain set 2.38 sele (resnam BGLY .and. type CA) end
            #        ! for GLY only
            if atom.name == 'CA':
                if atom.residue.name == 'GLY':
                    atom.solvent_radius = 2.38
                else:
                    atom.solvent_radius = 2.86
            # scalar wmain set 2.38 sele type CB2 end
            if atom.name == 'CB2':
                atom.solvent_radius = 2.38

            # !Hydrogens
            # scalar wmain set 0.00 sele type H* end
            #        ! for all hydrogens
            if atom.name.startswith('H'):
                atom.solvent_radius = 0.00

            # !Carbons
            # scalar wmain set 2.67 sele type CB end
            #        ! for all residues
            if atom.name == 'CB':
                atom.solvent_radius = 2.67
            # scalar wmain set 2.46 sele type CG* end
            #        ! for ARG, GLN, ILE, LYS, MET, PHE, THR, TRP, VAL,
            #              HSP, HSD, HSE
            if atom.residue.name in ['ARG', 'GLN', 'ILE', 'LYS', 'MET', 'PHE', 'THR', 'TRP', 'VAL',
                                     'HSP', 'HSD', 'HSE', 'LEU']:
                if atom.name.startswith('CG'):
                    atom.solvent_radius = 2.46
            # scalar wmain set 2.77 sele resnam GLU .and. type CG end
            # scalar wmain set 2.77 sele resnam BGLU .and. type CG end
            #        ! for GLU only
            if atom.residue.name == 'GLU':
                if atom.name == 'CG':
                    atom.solvent_radius = 2.77
            # scalar wmain set 2.44 sele type CD* end
            #        ! for ARG, ILE, LEU, LYS
            if atom.residue.name in ['ARG', 'ILE', 'LEU', 'LYS']:
                if atom.name.startswith('CD'):
                    atom.solvent_radius = 2.44
            # scalar wmain set 1.98 sele (resnam GLN .and. type CD) .or. (resnam ASN .and. type CG) .or. -
            #                            (resnam GLU .and. type CD) .or. (resnam ASP .and. type CG) end
            # scalar wmain set 1.98 sele (resnam BGLN .and. type CD) .or. (resnam BASN .and. type CG) .or. -
            #                            (resnam BGLU .and. type CD) .or. (resnam BASP .and. type CG) end
            #        ! for ASP, GLU, ASN, GLN
            if atom.residue.name in ['GLU', 'GLN']:
                if atom.name == 'CD':
                    atom.solvent_radius = 1.98
            if atom.residue.name in ['ASP', 'ASN']:
                if atom.name == 'CG':
                    atom.solvent_radius = 1.98
            # scalar wmain set 1.98 sele (resnam PRO .and. (type CB .or. type CG .or. type CD)) end
            # scalar wmain set 1.98 sele (resnam BPRO .and. (type CB .or. type CG .or. type CD)) end
            #        ! for PRO only
            if atom.residue.name == 'PRO':
                if atom.name in ['CB', 'CG', 'CD']:
                    atom.solvent_radius = 1.98
            # scalar wmain set 2.00 sele (resnam TYR .and. (type CE* .or. type CD* .or. -
            #                           type CZ)) .or. (resnam PHE .and. (type CE* .or. -
            #                           type CD* .or. type CZ))  end
            # scalar wmain set 2.00 sele (resnam BTYR .and. (type CE* .or. type CD* .or. -
            #                           type CZ)) .or. (resnam BPHE .and. (type CE* .or. -
            #                           type CD* .or. type CZ))  end
            #        ! for TYR, PHE rings
            if atom.residue.name in ['TYR', 'PHE']:
                if atom.name.startswith('CE') or atom.name.startswith('CD') or atom.name.startswith('CG') or \
                        atom.name == 'CZ':
                    atom.solvent_radius = 2.00
            # scalar wmain set 1.78 sele (resnam TRP .and. (type CE* .or. type CD* .or. -
            #                           type CZ* .or. type CH2)) end
            # scalar wmain set 1.78 sele (resnam BTRP .and. (type CE* .or. type CD* .or. -
            #                           type CZ* .or. type CH2)) end
            #        ! for TRP ring only
            if atom.residue.name == 'TRP':
                if atom.name.startswith('CE') or atom.name.startswith('CD') or atom.name.startswith('CZ') or \
                        atom.name == 'CH2':
                    atom.solvent_radius = 1.78
            # scalar wmain set 2.10 sele type CE end
            #        ! for MET only
            if atom.residue.name == 'MET':
                if atom.name == 'CE':
                    atom.solvent_radius = 2.10
            # scalar wmain set 2.80 sele (resnam ARG .and. type CZ) .or. (resnam LYS .and. type CE) end
            # scalar wmain set 2.80 sele (resnam BARG .and. type CZ) .or. (resnam BLYS .and. type CE) end
            # scalar wmain set 2.80 sele (resnam BORN .and. type CE) end
            #        ! for ARG, LYS
            if atom.residue.name in ['ARG', 'LYS', 'ORN']:
                if atom.name == 'CE' or atom.name == 'CZ':
                    atom.solvent_radius = 2.80
            # scalar wmain set 1.98 select (( resnam HSD .or. resnam HSE .or. resnam HSP) .and. type CE1) -
            #        .or. (( resnam HSD .or. resnam HSE .or. resnam HSP)  .and. type CD2) end
            # scalar wmain set 1.98 select (( resnam BHSD .or. resnam BHSE .or. resnam BHSP) .and. type CE1) -
            #        .or. (( resnam BHSD .or. resnam BHSE .or. resnam BHSP)  .and. type CD2) end
            #        ! for neutral HSD and protonated HSP
            if atom.residue.name in ['HSD', 'HSE', 'HSP']:
                if atom.name == 'CE1' or atom.name == 'CD2':
                    atom.solvent_radius = 1.98

            # !Oxygens
            # scalar wmain set 1.40 sele (resnam GLU .or. resnam ASP) .and. (type OE* .or. type OD*) end
            # scalar wmain set 1.40 sele (resnam BGLU .or. resnam BASP) .and. (type OE* .or. type OD*) end
            #        ! for GLU, ASP
            if atom.residue.name in ['GLU', 'ASP']:
                if atom.name.startswith('OE') or atom.name.startswith('OD'):
                    atom.solvent_radius = 1.40
            # scalar wmain set 1.42 sele (resnam ASN .or. resnam GLN) .and. (type OE* .or. type OD*) end
            # scalar wmain set 1.42 sele (resnam BASN .or. resnam BGLN) .and. (type OE* .or. type OD*) end
            #        ! for ASN, GLN
            if atom.residue.name in ['ASN', 'GLN']:
                if atom.name.startswith('OE') or atom.name.startswith('OD'):
                    atom.solvent_radius = 1.42
            # scalar wmain set 1.64 sele type OG* end
            #        ! for SER, THR
            if atom.residue.name in ['SER', 'THR']:
                if atom.name.startswith('OG'):
                    atom.solvent_radius = 1.64
            # scalar wmain set 1.85 sele (resnam TYR .and. type OH) end
            # scalar wmain set 1.85 sele (resnam BTYR .and. type OH) end
            #        ! for TYR only
            if atom.residue.name == 'TYR':
                if atom.name == 'OH':
                    atom.solvent_radius = 1.85
            # scalar wmain set 2.2 select resname TIP3 .and. type OH2 end
            #        ! for explicit water molecules
            if atom.residue.name == 'TIP3':
                if atom.name == 'OH2':
                    atom.solvent_radius = 2.2

            # !Nitrogens
            # scalar wmain set 1.80 sele resnam HSD  .and. (type NE2 .or. type ND1) end
            # scalar wmain set 1.80 sele resnam BHSD  .and. (type NE2 .or. type ND1) end
            #        ! for neutral HSD
            # scalar wmain set 1.80 sele resnam HSE  .and. (type NE2 .or. type ND1) end
            # scalar wmain set 1.80 sele resnam BHSE  .and. (type NE2 .or. type ND1) end
            #        ! for neutral HSE
            if atom.residue.name in ['HSD', 'HSE']:
                if atom.name == 'NE2' or atom.name == 'ND1':
                    atom.solvent_radius = 1.80
            # scalar wmain set 2.30 sele resnam HSP  .and. (type NE2 .or. type ND1) end
            # scalar wmain set 2.30 sele resnam BHSP  .and. (type NE2 .or. type ND1) end
            #        ! for protonated HSP
            if atom.residue.name == 'HSP':
                if atom.name == 'NE2' or atom.name == 'ND1':
                    atom.solvent_radius = 2.30
            # scalar wmain set 2.13 sele resnam ARG .and. (type NH* .or. type NE) .or. -
            #                       (resnam LYS .and. type NZ) end
            # scalar wmain set 2.13 sele resnam BARG .and. (type NH* .or. type NE) .or. -
            #                       (resnam BLYS .and. type NZ) end
            # scalar wmain set 2.13 sele (resnam BORN .and. type NZ) end
            #        ! for ARG, LYS
            if atom.residue.name in ['ARG', 'LYS', 'ORN']:
                if atom.name.startswith('NH') or atom.name == 'NE' or atom.name == 'NZ':
                    atom.solvent_radius = 2.13
            # scalar wmain set 2.15 sele (resnam GLN .and. type NE2) .or. (resnam ASN .and. type ND2) end
            # scalar wmain set 2.15 sele (resnam BGLN .and. type NE2) .or. (resnam BASN .and. type ND2) end
            #        ! for GLN, ASN
            if atom.residue.name in ['GLN', 'ASN']:
                if atom.name == 'NE2' or atom.name == 'ND2':
                    atom.solvent_radius = 2.15
            # scalar wmain set 2.40 sele resnam TRP .and. type NE1 end
            # scalar wmain set 2.40 sele resnam BTRP .and. type NE1 end
            #        ! for TRP
            if atom.residue.name == 'TRP':
                if atom.name == 'NE1':
                    atom.solvent_radius = 2.40

            # !Sulphur
            # scalar wmain set 2.00 sele type S* end
            #        ! for MET, CYS
            if atom.residue.name in ['MET', 'CYS']:
                if atom.name.startswith('S'):
                    atom.solvent_radius = 2.00

            # ! Phosphate atoms in phosphotyrosine, phosphoserine and phosphothreonine
            # calc radius = 2.35*@factor
            # scalar wmain set @radius sele (resn tyr .and. type p1) end
            # scalar wmain set @radius sele (resn ser .and. type p ) end
            if atom.residue.name in ['SER', 'TYR']:
                if atom.name == 'p1' or atom.name == 'p':
                    atom.solvent_radius = 2.35
            # calc radius = 1.49*@factor
            # scalar wmain set @radius sele (resn tyr .and. (type o2  .or. type o3  .or. type o4)) end
            # scalar wmain set @radius sele (resn ser .and. (type o1p .or. type o2p .or. type o3p .or. type ot )) end
            if atom.residue.name == 'TYR':
                if atom.name == 'o2' or atom.name == 'o3' or atom.name == 'o4':
                    atom.solvent_radius = 1.49
            if atom.residue.name in ['SER', 'THR']:
                if atom.name == 'o1p' or atom.name == 'o2p' or atom.name == 'o3p' or atom.name == 'ot':
                    atom.solvent_radius = 1.49

        # NUCLEIC ACID ATOMS
        elif atom.residue.name in ['ADE', 'GUA' , 'CYT', 'THY', 'URA', 'DMPA', 'ATP', 'ADP']:
            factor = 1
            # if @?factor eq 0 set factor = 1
            #
            # ! Set all H radii to 0.0
            # calc radius = 0.0*@factor
            # scalar wmain set @radius sele chem H* end
            if atom.name.startswith('H'):
                atom.solvent_radius = 0.00

            # ! Purine base atoms
            # calc radius = 1.75*@factor
            # scalar wmain set @radius sele resn ade .and. type n1 end
            if atom.residue.name == 'ADE':
                if atom.name == 'N1':
                    atom.solvent_radius = 1.75 * factor
            # calc radius = 2.17*@factor
            # scalar wmain set @radius sele resn ade .and. type n6 end
                elif atom.name == 'N6':
                    atom.solvent_radius = 2.17 * factor
            # calc radius = 2.15*@factor
            # scalar wmain set @radius sele resn gua .and. type n1 end
            if atom.residue.name == 'GUA':
                if atom.name == 'N1':
                    atom.solvent_radius = 2.15 * factor
            # calc radius = 1.55*@factor
            # scalar wmain set @radius sele resn gua .and. type o6 end
                elif atom.name == 'O6':
                    atom.solvent_radius = 1.55 * factor
            # calc radius = 2.12*@factor
            # scalar wmain set @radius sele resn gua .and. type n2 end
                elif atom.name == 'N2':
                    atom.solvent_radius = 2.12 * factor
            # calc radius = 2.15*@factor
            # scalar wmain set @radius sele (resn gua .or. resn ade) .and. type c2 end
            if atom.residue.name in ['GUA', 'ADE']:
                if atom.name == 'C2':
                    atom.solvent_radius = 2.15 * factor
            # calc radius = 1.69*@factor
            # scalar wmain set @radius sele (resn gua .or. resn ade) .and. type n3 end
                elif atom.name == 'N3':
                    atom.solvent_radius = 1.69 * factor
            # calc radius = 2.12*@factor
            # scalar wmain set @radius sele (resn gua .or. resn ade) .and. type c4 end
                elif atom.name == 'C4':
                    atom.solvent_radius = 2.12 * factor
            # calc radius = 2.12*@factor
            # scalar wmain set @radius sele (resn gua .or. resn ade) .and. type c5 end
                elif atom.name == 'C5':
                    atom.solvent_radius = 2.12 * factor
            # calc radius = 2.12*@factor
            # scalar wmain set @radius sele (resn gua .or. resn ade) .and. type c6 end
                elif atom.name == 'C6':
                    atom.solvent_radius = 2.12 * factor
            # calc radius = 1.69*@factor
            # scalar wmain set @radius sele (resn gua .or. resn ade) .and. type n7 end
                elif atom.name == 'N7':
                    atom.solvent_radius = 1.69 * factor
            # calc radius = 2.12*@factor
            # scalar wmain set @radius sele (resn gua .or. resn ade) .and. type c8 end
                elif atom.name == 'C8':
                    atom.solvent_radius = 2.12 * factor
            # calc radius = 2.13*@factor
            # scalar wmain set @radius sele (resn gua .or. resn ade) .and. type n9 end
                elif atom.name == 'N9':
                    atom.solvent_radius = 2.13 * factor
            # calc radius = 2.30*@factor
            # scalar wmain set @radius sele (resn gua .or. resn ade) .and. type c9 end
                elif atom.name == 'C9':
                    atom.solvent_radius = 2.30 * factor

            # ! Pyrimidine base atoms
            # calc radius = 2.20*@factor
            # scalar wmain set @radius sele (resn cyt .or. resn thy .or. resn ura) .and. type n1 end
            if atom.residue.name in ['CYT', 'THY', 'URA']:
                if atom.name == 'N1':
                    atom.solvent_radius = 2.20 * factor
            # calc radius = 2.04*@factor
            # scalar wmain set @radius sele (resn cyt .or. resn thy .or. resn ura) .and. type c2 end
                elif atom.name == 'C2':
                    atom.solvent_radius = 2.04 * factor
            # calc radius = 2.12*@factor
            # scalar wmain set @radius sele (resn cyt .or. resn thy .or. resn ura) .and. type c4 end
                elif atom.name == 'C4':
                    atom.solvent_radius = 2.12 * factor
            # calc radius = 2.25*@factor
            # scalar wmain set @radius sele (resn cyt .or. resn thy .or. resn ura) .and. type c5 end
                elif atom.name == 'C5':
                    atom.solvent_radius = 2.25 * factor
            # calc radius = 2.25*@factor
            # scalar wmain set @radius sele (resn cyt .or. resn thy .or. resn ura) .and. type c6 end
                elif atom.name == 'C6':
                    atom.solvent_radius = 2.25 * factor
            # calc radius = 1.60*@factor
            # scalar wmain set @radius sele (resn cyt .or. resn thy .or. resn ura) .and. type o2 end
                elif atom.name == 'O2':
                    atom.solvent_radius = 1.60 * factor
            # calc radius = 2.30*@factor
            # scalar wmain set @radius sele (resn cyt .or. resn thy .or. resn ura) .and. type c1 end
                elif atom.name == 'C1':
                    atom.solvent_radius = 2.30 * factor
            # calc radius = 1.68*@factor
            # scalar wmain set @radius sele resn cyt .and. type n3 end
            if atom.residue.name == 'CYT':
                if atom.name == 'N3':
                    atom.solvent_radius = 1.68 * factor
            # calc radius = 2.08*@factor
            # scalar wmain set @radius sele resn cyt .and. type n4 end
                elif atom.name == 'N4':
                    atom.solvent_radius = 2.08 * factor
            # calc radius = 2.20*@factor
            # scalar wmain set @radius sele (resn thy .or. resn ura) .and. type n3 end
            if atom.residue.name in ['THY', 'URA']:
                if atom.name == 'N3':
                    atom.solvent_radius = 2.20 * factor
            # calc radius = 1.60*@factor
            # scalar wmain set @radius sele (resn thy .or. resn ura) .and. type o4 end
                elif atom.name == 'O4':
                    atom.solvent_radius = 1.60 * factor
            # calc radius = 2.30*@factor
            # scalar wmain set @radius sele resn thy .and. type c5m end
            if atom.residue.name == 'THY':
                if atom.name == 'C5M':
                    atom.solvent_radius = 2.30 * factor

            # ! sugar atoms
            # calc radius = 2.57*@factor
            # scalar wmain set @radius sele type c1' end
            if atom.name == "C1'":
                atom.solvent_radius = 2.57 * factor
            # calc radius = 2.70*@factor
            # scalar wmain set @radius sele type c2' end
            if atom.name == "C2'":
                atom.solvent_radius = 2.70 * factor
            # calc radius = 2.73*@factor
            # scalar wmain set @radius sele type c3' end
            if atom.name == "C3'":
                atom.solvent_radius = 2.73 * factor
            # calc radius = 2.50*@factor
            # scalar wmain set @radius sele type c4' end
            if atom.name == "C4'":
                atom.solvent_radius = 2.50 * factor
            # calc radius = 1.55*@factor
            # scalar wmain set @radius sele type o4' end
            if atom.name == "O4'":
                atom.solvent_radius = 1.55 * factor
            # calc radius = 2.57*@factor
            # scalar wmain set @radius sele type c5' end
            if atom.name == "C5'":
                atom.solvent_radius = 2.57 * factor
            # calc radius = 1.75*@factor
            # scalar wmain set @radius sele type o2' end
            if atom.name == "O2'":
                atom.solvent_radius = 1.75 * factor

            if atom.name in ["O3'", "O5'"]:
                bondeds = list(atom.bond_partners)
                # ! add radii for blocking group hydroxyl oxygens
                # define oter  sele (type o3' .or. type o5') .and. .bonded. (type h3t .or. type h5t) end
                # calc radius = 1.72*@factor
                # scalar wmain set @radius sele oter end
                if "H3T" in [at.name for at in bondeds] or "H5T" in [at.name for at in bondeds]:
                    atom.solvent_radius = 1.72 * factor
                else:
                    # calc radius = 1.65*@factor
                    # scalar wmain set @radius sele type o3' end
                    # calc radius = 1.65*@factor
                    # scalar wmain set @radius sele type o5' end
                    atom.solvent_radius = 1.65 * factor

            # ! atoms for sugar2phos
            # calc radius = 1.65*@factor
            # scalar wmain set @radius sele resn ade .and. type o5t  end
            if atom.residue.name in ['ADE', 'GUA', 'CYT', 'THY', 'URA']:
                if atom.name == 'O5T':
                    atom.solvent_radius = 1.65 * factor
            # calc radius = 2.30*@factor
            # scalar wmain set @radius sele resn ade .and. type c5t  end
                elif atom.name == 'C5T':
                    atom.solvent_radius = 2.30 * factor
            # calc radius = 2.35*@factor
            # scalar wmain set @radius sele resn ade .and. type p3   end
                elif atom.name == 'P3':
                    atom.solvent_radius = 2.35 * factor
            # calc radius = 1.49*@factor
            # scalar wmain set @radius sele resn ade .and. type o1p3 end
                elif atom.name == 'O1P3':
                    atom.solvent_radius = 1.49 * factor
            # calc radius = 1.49*@factor
            # scalar wmain set @radius sele resn ade .and. type o2p3 end
                elif atom.name == 'O2P3':
                    atom.solvent_radius = 1.49 * factor
            # calc radius = 1.65*@factor
            # scalar wmain set @radius sele resn ade .and. type o3t  end
                elif atom.name == 'O3T':
                    atom.solvent_radius = 1.65 * factor
            # calc radius = 2.30*@factor
            # scalar wmain set @radius sele resn ade .and. type c3t  end
                elif atom.name == 'C3T':
                    atom.solvent_radius = 2.30 * factor

            # ! phosphate atoms
            # calc radius = 2.35*@factor
            # scalar wmain set @radius sele (resn gua .or. resn ade .or. resn cyt -
            #                     .or. resn thy .or. resn ura) .and. type p end
            if atom.residue.name in ['GUA', 'ADE', 'CYT', 'THY', 'URA']:
                if atom.name == 'P':
                    atom.solvent_radius = 2.35 * factor
            # calc radius = 1.49*@factor
            # scalar wmain set @radius sele (resn gua .or. resn ade .or. resn cyt -
            #                     .or. resn thy .or. resn ura) .and. type o1p end
                elif atom.name == 'O1P':
                    atom.solvent_radius = 1.49 * factor
            # calc radius = 1.49*@factor
            # scalar wmain set @radius sele (resn gua .or. resn ade .or. resn cyt -
            #                     .or. resn thy .or. resn ura) .and. type o2p end
                elif atom.name == 'O2P':
                    atom.solvent_radius = 1.49 * factor

            # ! phosphate atoms in ADP or ATP
            # calc radius = 2.35*@factor
            # scalar wmain set @radius sele type pa .or. type pb .or. type pg end
            if atom.name == 'PA' or atom.name == 'PB' or atom.name == 'PG':
                atom.solvent_radius = 2.35 * factor
            # calc radius = 1.65*@factor
            # scalar wmain set @radius sele (resn atp .and. ( type o3a .or. type o3b)) .or. -
            #                               (resn adp .and. ( type o3a)) end
            # calc radius = 1.49*@factor
            # scalar wmain set @radius sele type o1a .or. type o2a .or. type o1b .or. type o2b -
            #                  .or. type o1g .or. type o2g .or. type o3g .or. (resn adp .and. type o3b) end
            if atom.residue.name == 'ATP':
                if atom.name == 'O3A' or atom.name == 'O3B':
                    atom.solvent_radius = 1.65 * factor

            if atom.residue.name == 'ADP':
                if atom.name == 'O3A':
                    atom.solvent_radius = 1.65 * factor
                elif atom.name == 'O3B':
                    atom.solvent_radius = 1.49 * factor

            if atom.name in ['O1A', 'O2A', 'O1B', 'O2B', 'O1G', 'O2G', 'O3G']:
                atom.solvent_radius = 1.49 * factor

        # OTHER ATOMS
        else:
            # ! Set to zero all H radii
            # scalar wmain set 0.0 sele chem H* end
            if atom.name.startswith('H'):
                atom.solvent_radius = 0.00

            # !Carbons
            elif atom.name.startswith('C'):
                # from Table S5 (https://pubs.acs.org/doi/10.1021/acs.jcim.1c00177)
                if atom.type in [x[:4] for x in
                                 ['CG1N1', 'CG2D1', 'CG2D1O', 'CG2D2', 'CG2DC1', 'CG2O1', 'CG2O2',
                                  'CG2O4', 'CG2R51', 'CG2R53', 'CG2R61', 'CG2R62', 'CG2R63', 'CG2R67',
                                  'CG301', 'CG302', 'CG311', 'CG312', 'CG321', 'CG322', 'CG331',
                                  'CG3C50', 'CG3C51', 'CG3C52', 'CG3RC1']]:
                    atom.solvent_radius = 2.46
                # from Table S5 (https://pubs.acs.org/doi/10.1021/acs.jcim.1c00177)
                elif atom.type == 'CG2RC0'[:4]:
                    atom.solvent_radius = 2.12
                else:
                    # scalar wmain set 2.3 sele chem C* end
                    atom.solvent_radius = 2.30

            # !Oxygens
            elif atom.name.startswith('O'):
                # from Table S5 (https://pubs.acs.org/doi/10.1021/acs.jcim.1c00177)
                if atom.type in [x[:4] for x in ['OG2D1', 'OG2N1', 'OG301', 'OG302', 'OG303', 'OG311', 'OG312',
                                                 'OG3R60']]:
                    atom.solvent_radius = 1.64
                # from Table S5 (https://pubs.acs.org/doi/10.1021/acs.jcim.1c00177)
                elif atom.type == 'OG2D4'[:4]:
                    atom.solvent_radius = 1.60
                # from Table S5 (https://pubs.acs.org/doi/10.1021/acs.jcim.1c00177)
                elif atom.type == 'OG2P1'[:4]:
                    atom.solvent_radius = 1.49
                else:
                    # scalar wmain set 1.8 sele chem O* end
                    atom.solvent_radius = 1.8

            # !Nitrogens
            elif atom.name.startswith('N'):
                # from Table S5 (https://pubs.acs.org/doi/10.1021/acs.jcim.1c00177)
                if atom.type == 'NG2S1'[:4]:
                    atom.solvent_radius = 2.23
                else:
                    # scalar wmain set 2.3 sele chem N* end
                    atom.solvent_radius = 2.30

            # !Sulphur
            # scalar wmain set 2.3 sele chem S* end
            elif atom.name.startswith('S'):
                atom.solvent_radius = 2.30

            # !Phosphorus
            elif atom.name.startswith('P'):
                # from Table S5 (https://pubs.acs.org/doi/10.1021/acs.jcim.1c00177)
                atom.solvent_radius = 2.35

            # !Halogens rstd from Table S8 (https://pubs.acs.org/doi/10.1021/acs.jcim.1c00177)
            elif atom.name.startswith('CL'):
                atom.solvent_radius = 1.86
            elif atom.name.startswith('BR'):
                atom.solvent_radius = 1.98
            elif atom.name.startswith('I'):
                atom.solvent_radius = 2.24

            else:
                atom.solvent_radius = 2.00

    try:
        parm.parm_data['RADIUS_SET'][0] = \
                'charmm radii (charmm_radii)'
    except AttributeError:
        pass
    _screen1(parm)

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def _screen1(parm):
    """ Applies the first set of screening parameters found in tleap source """
    for i, atom in enumerate(parm.atoms):
        if atom.atomic_number == 1:
            atom.screen = 0.85
        elif atom.atomic_number == 6:
            atom.screen = 0.72
        elif atom.atomic_number == 7:
            atom.screen = 0.79
        elif atom.atomic_number == 8:
            atom.screen = 0.85
        elif atom.atomic_number == 9:
            atom.screen = 0.88
        elif atom.atomic_number == 15:
            atom.screen = 0.86
        elif atom.atomic_number == 16:
            atom.screen = 0.96
        else:
            atom.screen = 0.8

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def _screen2(parm):
    """
    Applies the second set of screening parameters found in tleap source
    (unused as far as I can tell)
    """
    for i, atom in enumerate(parm.atoms):
        if atom.atomic_number == 1:
            atom.screen = 0.8461
        elif atom.atomic_number == 6:
            atom.screen = 0.9615
        elif atom.atomic_number == 7:
            atom.screen = 0.9343
        elif atom.atomic_number == 8:
            atom.screen = 1.0088
        elif atom.atomic_number == 11:
            atom.screen = 1.0000
        elif atom.atomic_number == 12:
            atom.screen = 1.0000
        elif atom.atomic_number == 15:
            atom.screen = 1.0700
        elif atom.atomic_number == 16:
            atom.screen = 1.1733
        else:
            atom.screen = 0.8

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def _screen3(parm):
    """
    Applies the third and final set of screening parameters found in tleap
    source (unused as far as I can tell)
    """
    for i, atom in enumerate(parm.atoms):
        if atom.atomic_number == 1:
            atom.screen = 0.8846
        elif atom.atomic_number == 6:
            atom.screen = 0.9186
        elif atom.atomic_number == 7:
            atom.screen = 0.8733
        elif atom.atomic_number == 8:
            atom.screen = 0.8836
        elif atom.atomic_number == 11:
            atom.screen = 1.0000
        elif atom.atomic_number == 12:
            atom.screen = 1.0000
        elif atom.atomic_number == 15:
            atom.screen = 0.9604
        elif atom.atomic_number == 16:
            atom.screen = 0.9323
        else:
            atom.screen = 0.8

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

_call_method = dict(bondi=bondi, mbondi=mbondi, mbondi_pb2=mbondi_pb2, mbondi_pb3=mbondi_pb3, mbondi2=mbondi2,
                    mbondi3=mbondi3, amber6=amber6, charmm_radii=charmm_radii)

def ChRad(parm, radii_set):
    if radii_set not in _call_method:
        raise ChangeRadiiError(f"You must choose from {list(_call_method.keys())} radii sets")
    _call_method[radii_set](parm)



"""
Tool for parsing and writing OFF library files to and from dictionaries of
ResidueTemplate objects
"""
from collections import OrderedDict
from contextlib import closing
import numpy as np
from ..topologyobjects import Atom
from ..constants import RAD_TO_DEG
from ..exceptions import AmberWarning
from ..formats.registry import FileFormatType
from ..modeller.residue import ResidueTemplate, ResidueTemplateContainer
from ..modeller.residue import PROTEIN, NUCLEIC, SOLVENT, UNKNOWN
from .. import periodic_table as pt
from ..utils.io import genopen
import re
import warnings

class AmberOFFLibrary(metaclass=FileFormatType):
    """
    Class containing static methods responsible for parsing and writing OFF
    libraries
    """
    #===================================================

    # Useful regexes
    _headerre = re.compile(r'!!index *array *str')
    _resre = re.compile(r'\s*"(\S*?)"\s*$')
    _sec1re = re.compile(r'!entry\.(\S*?)\.unit\.atoms *table *str *name *str'
                         r' *type *int *typex *int *resx *int *flags *int'
                         r' *seq *int *elmnt *dbl *chg')
    _sec2re = re.compile(r'!entry\.(\S*?)\.unit\.atomspertinfo *table *str'
                         r' *pname *str *ptype *int *ptypex *int *pelmnt'
                         r' *dbl *pchg')
    _sec3re = re.compile(r'!entry\.(\S*?)\.unit\.boundbox *array *dbl')
    _sec4re = re.compile(r'!entry\.(\S*?)\.unit\.childsequence *single *int')
    _sec5re = re.compile(r'!entry\.(\S*?)\.unit\.connect *array *int')
    _sec6re = re.compile(r'!entry\.(\S*?)\.unit\.connectivity *table *int'
                         r' *atom1x *int *atom2x *int *flags')
    _sec7re = re.compile(r'!entry\.(\S*?)\.unit\.hierarchy *table *str'
                         r' *abovetype *int *abovex *str *belowtype *int'
                         r' *belowx')
    _sec8re = re.compile(r'!entry\.(\S*?)\.unit\.name *single *str')
    _sec9re = re.compile(r'!entry\.(\S*?)\.unit\.positions *table *dbl *x'
                         r' *dbl *y *dbl *z')
    _sec10re = re.compile(r'!entry\.(\S*?)\.unit\.residueconnect *table'
                          r' *int *c1x *int *c2x *int *c3x *int *c4x *int'
                          r' *c5x *int *c6x')
    _sec11re = re.compile(r'!entry\.(\S*?)\.unit\.residues *table *str *name'
                          r' *int *seq *int *childseq *int *startatomx *str'
                          r' *restype *int *imagingx')
    _sec12re = re.compile(r'!entry\.(\S*?)\.unit\.residuesPdbSequenceNumber'
                          r' *array *int')
    _sec13re = re.compile(r'!entry\.(\S*?)\.unit\.solventcap *array *dbl')
    _sec14re = re.compile(r'!entry\.(\S*?)\.unit\.velocities *table *dbl *x'
                          r' *dbl *y *dbl *z')

    #===================================================

    @classmethod
    def id_format(cls, filename):
        """ Sees if an open file is an OFF library file.

        Parameters
        ----------
        filename : str
            The name of the file to see if it is an OFF file format

        Returns
        -------
        is_fmt : bool
            True if it is recognized as OFF, False otherwise
        """
        with closing(genopen(filename, 'r')) as f:
            if cls._headerre.match(f.readline()):
                return True
            return False

    #===================================================

    @classmethod
    def parse(cls, filename):
        """ Parses an Amber OFF library

        Parameters
        ----------
        filename : str or file-like iterable
            The file name or file object to parse. If it is an iterable, it will
            be exhausted

        Returns
        -------
        residues : OrderedDict {str : :class:`ResidueTemplate`}
            Dictionary pairing residue names with their :class:`ResidueTemplate`
            objects

        Raises
        ------
        ValueError if the first line does not match the file format. This line
        will be consumed

        IOError if filename is the name of a file that does not exist

        RuntimeError if EOF is reached prematurely or other formatting issues
        found
        """
        if isinstance(filename, str):
            fileobj = genopen(filename, 'r')
            own_handle = True
        else:
            fileobj = filename
            own_handle = False
        # Now parse the library file
        line = fileobj.readline()
        if not cls._headerre.match(line):
            raise ValueError('Unrecognized OFF file format')
        # Build the return value
        residues = OrderedDict()
        # Pull a list of all the residues we expect to find
        line = fileobj.readline()
        rematch = cls._resre.match(line)
        while rematch and line:
            name = rematch.groups()[0]
            residues[name] = None
            line = fileobj.readline()
            rematch = cls._resre.match(line)
        if not line:
            raise RuntimeError('Unexpected EOF in Amber OFF library')
        # Now make sure we have the next expected line
        while line:
            if not line.strip():
                line = fileobj.readline()
                continue
            rematch = cls._sec1re.match(line)
            if not rematch:
                raise RuntimeError('Expected atoms table not found')
            name = rematch.groups()[0]
            residues[name] = cls._parse_residue(fileobj, name)
            line = fileobj.readline()

        if own_handle:
            fileobj.close()

        return residues

    #===================================================

    @classmethod
    def _parse_residue(cls, fileobj, name):
        """
        Parses the residue information out of the OFF file assuming the file
        is pointed at the first line of an atoms table section of the OFF file

        Parameters
        ----------
        fileobj : file-like
            Assumed to be open for read, this file is parsed until the *next*
            atom table is read
        name : str
            The name of the residue being processed right now
        """
        container = ResidueTemplateContainer(name)
        nres = 1
        templ = ResidueTemplate(name)
        line = fileobj.readline()
        while line[0] != '!':
            nam, typ, typx, resx, flags, seq, elmnt, chg = line.split()
            nam = _strip_enveloping_quotes(nam)
            typ = _strip_enveloping_quotes(typ)
            typx = int(typx)
            resx = int(resx)
            flags = int(flags)
            seq = int(seq)
            elmnt = int(elmnt)
            chg = float(chg)
            atom = Atom(atomic_number=elmnt, type=typ, name=nam, charge=chg)
            if resx == nres + 1:
                container.append(templ)
                nres += 1
                templ = ResidueTemplate(name)
            templ.add_atom(atom)
            line = fileobj.readline()
            # Skip blank lines
            while line and not line.strip():
                line = fileobj.readline()
        container.append(templ)
        if nres > 1:
            start_atoms = []
            runsum = 0
            for res in container:
                start_atoms.append(runsum)
                runsum += len(res)
        # Make sure we get the next section
        rematch = cls._sec2re.match(line)
        if not rematch:
            raise RuntimeError('Expected pertinfo table not found')
        elif rematch.groups()[0] != name:
            raise RuntimeError(
                f'Found residue {rematch.groups()[0]} while processing residue {name}'
            )
        line = fileobj.readline()
        while line[0] != '!':
            if not line:
                raise RuntimeError('Unexpected EOF in Amber OFF library')
            # Not used, just skip
            # TODO sanity check
            line = fileobj.readline()
        rematch = cls._sec3re.match(line)
        if not rematch:
            raise RuntimeError('Expected boundbox table not found')
        elif rematch.groups()[0] != name:
            raise RuntimeError(
                f'Found residue {rematch.groups()[0]} while processing residue {name}'
            )
        # Only 5 lines
        try:
            hasbox = float(fileobj.readline().strip())
            angle = float(fileobj.readline().strip())
            a = float(fileobj.readline().strip())
            b = float(fileobj.readline().strip())
            c = float(fileobj.readline().strip())
        except ValueError:
            raise RuntimeError('Error processing boundbox table entries')
        else:
            if hasbox > 0:
                if angle < 3.15:
                    # No box is this acute -- must be in radians
                    angle *= RAD_TO_DEG
                container.box = [a, b, c, angle, angle, angle]
        # Get the child sequence entry
        line = fileobj.readline()
        rematch = cls._sec4re.match(line)
        if not rematch:
            raise RuntimeError('Expected childsequence table not found')
        elif rematch.groups()[0] != name:
            raise RuntimeError(
                f"Found residue {rematch.groups()[0]} while processing residue {name}"
            )
        n = int(fileobj.readline().strip())
        if nres + 1 != n:
            warnings.warn(
                f"Unexpected childsequence ({n}); expected {nres+1} for residue {name}",
                AmberWarning,
            )
        elif not isinstance(templ, ResidueTemplate) and n != len(templ) + 1:
            raise RuntimeError("child sequence must be # of residues in the unit + 1")
        # Get the CONNECT array to set head and tail
        line = fileobj.readline()
        rematch = cls._sec5re.match(line)
        if not rematch:
            raise RuntimeError('Expected connect array not found')
        elif rematch.groups()[0] != name:
            raise RuntimeError(
                f"Found residue {rematch.groups()[0]} while processing residue {name}"
            )
        try:
            head = int(fileobj.readline().strip())
            tail = int(fileobj.readline().strip())
        except ValueError:
            raise RuntimeError('Error processing connect table entries')
        if head > 0 and nres == 1:
            templ.head = templ[head-1]
        elif head > 0 and nres > 1:
            if head < sum((len(r) for r in container)):
                raise RuntimeError('HEAD on multi-residue unit not supported')
        if tail > 0 and nres == 1:
            templ.tail = templ[tail-1]
        elif tail > 0 and nres > 1:
            if tail < sum((len(r) for r in container)):
                warnings.warn(
                    f'TAIL on multi-residue unit not supported ({name}). Ignored...', AmberWarning
                )
        # Get the connectivity array to set bonds
        line = fileobj.readline()
        if len(templ.atoms) > 1:
            rematch = cls._sec6re.match(line)
            if not rematch:
                raise RuntimeError('Expected connectivity table not found')
            elif rematch.groups()[0] != name:
                raise RuntimeError(
                    f"Found residue {rematch.groups()[0]} while processing residue {name}"
                )
            line = fileobj.readline()
            while line[0] != '!':
                i, j, flag = line.split()
                line = fileobj.readline()
                if nres > 1:
                    # Find which residue we belong in
                    i = int(i) - 1
                    j = int(j) - 1
                    for ii, idx in enumerate(start_atoms):
                        if idx > i:
                            ii -= 1
                            break
                    start_idx = start_atoms[ii]
                    container[ii].add_bond(i-start_idx, j-start_idx)
                else:
                    templ.add_bond(int(i)-1, int(j)-1)
        # Get the hierarchy table
        rematch = cls._sec7re.match(line)
        if not rematch:
            raise RuntimeError('Expected hierarchy table not found')
        elif rematch.groups()[0] != name:
            raise RuntimeError(
                f"Found residue {rematch.groups()[0]} while processing residue {name}"
            )
        line = fileobj.readline()
        while line[0] != '!':
            # Skip this section... not used
            # TODO turn this into a sanity check
            line = fileobj.readline()
        # Get the unit name
        rematch = cls._sec8re.match(line)
        if not rematch:
            raise RuntimeError('Expected unit name string not found')
        elif rematch.groups()[0] != name:
            raise RuntimeError(
                f"Found residue {rematch.groups()[0]} while processing residue {name}"
            )
        fileobj.readline() # Skip this... not used
        line = fileobj.readline()
        # Get the atomic positions
        rematch = cls._sec9re.match(line)
        if not rematch:
            raise RuntimeError('Expected unit positions table not found')
        elif rematch.groups()[0] != name:
            raise RuntimeError(
                f"Found residue {rematch.groups()[0]} while processing residue {name}"
            )
        for res in container:
            for atom in res:
                x, y, z = fileobj.readline().split()
                atom.xx, atom.xy, atom.xz = float(x), float(y), float(z)
        line = fileobj.readline()
        # Get the residueconnect table
        rematch = cls._sec10re.match(line)
        if not rematch:
            raise RuntimeError('Expected unit residueconnect table not found')
        elif rematch.groups()[0] != name:
            raise RuntimeError(
                f"Found residue {rematch.groups()[0]} while processing residue {name}"
            )
        for i in range(nres):
            c1,c2,c3,c4,c5,c6 = (int(x) for x in fileobj.readline().split())
            if (c1 > 0 and templ.head is not None and
                    templ.head is not templ[c1-1]):
                raise RuntimeError('HEAD atom is not connect0')
            if (c2 > 0 and templ.tail is not None and
                    templ.tail is not templ[c2-1]):
                raise RuntimeError('TAIL atom is not connect1')
            for i in (c3, c4, c5, c6):
                if i == 0: continue
                templ.connections.append(templ[i-1])
        # Get the residues table
        line = fileobj.readline()
        rematch = cls._sec11re.match(line)
        if not rematch:
            raise RuntimeError('Expected unit residues table not found')
        elif rematch.groups()[0] != name:
            raise RuntimeError(
                f"Found residue {rematch.groups()[0]} while processing residue {name}"
            )
        for i in range(nres):
            resname, id, next, start, typ, img = fileobj.readline().split()
            resname = _strip_enveloping_quotes(resname)
            id = int(id)
            start = int(start)
            next = int(next)
            typ = _strip_enveloping_quotes(typ)
            img = int(img)
            if next - start != len(container[i]):
                warnings.warn(
                    f'residue table predicted {next - start}, not {len(container[i])} atoms for residue {name}',
                    AmberWarning,
                )
            if typ == 'p':
                container[i].type = PROTEIN
            elif typ == 'n':
                container[i].type = NUCLEIC
            elif typ == 'w':
                container[i].type = SOLVENT
            elif typ != '?':
                warnings.warn('Unknown residue type "%s"' % typ, AmberWarning)
            if nres > 1:
                container[i].name = resname
        # Get the residues sequence table
        line = fileobj.readline()
        rematch = cls._sec12re.match(line)
        if not rematch:
            raise RuntimeError('Expected residue sequence number not found')
        elif rematch.groups()[0] != name:
            raise RuntimeError(
                f"Found residue {rematch.groups()[0]} while processing residue {name}"
            )
        for i in range(nres):
            #TODO sanity check
            fileobj.readline()
        line = fileobj.readline()
        # Get the solventcap array
        rematch = cls._sec13re.match(line)
        if not rematch:
            raise RuntimeError('Expected unit solventcap array not found')
        elif rematch.groups()[0] != name:
            raise RuntimeError(
                f"Found residue {rematch.groups()[0]} while processing residue {name}"
            )
        # Ignore the solvent cap
        fileobj.readline()
        fileobj.readline()
        fileobj.readline()
        fileobj.readline()
        fileobj.readline()
        # Velocities
        line = fileobj.readline()
        rematch = cls._sec14re.match(line)
        if not rematch:
            raise RuntimeError('Expected unit solventcap array not found')
        elif rematch.groups()[0] != name:
            raise RuntimeError(
                f"Found residue {rematch.groups()[0]} while processing residue {name}"
            )
        for res in container:
            for atom in res:
                vx, vy, vz = (float(x) for x in fileobj.readline().split())
                atom.vx, atom.vy, atom.vz = vx, vy, vz

        if nres > 1:
            return container
        return templ

    #===================================================

    @classmethod
    def write(cls, lib, dest):
        """ Writes a dictionary of ResidueTemplate units to a file in OFF format

        Parameters
        ----------
        lib : dict {str : :class:`ResidueTemplate`}
            Items can be either :class:`ResidueTemplate` or
            :class:`ResidueTemplateContainer` instances
        dest : str or file-like
            Either a file name or a file-like object to write the file to
        """
        own_handle = False
        if not hasattr(dest, 'write'):
            dest = genopen(dest, 'w')
            own_handle = True
        # Write the residues in alphabetical order
        names = sorted(lib.keys())
        dest.write('!!index array str\n')
        for name in names:
            dest.write(' "%s"\n' % name)
        for name in names:
            cls._write_residue(dest, lib[name])

        if own_handle: dest.close()

    #===================================================

    @staticmethod
    def _write_residue(dest, res):
        """ Writes a residue to an open file handle

        Parameters
        ----------
        dest : file-like
            File object to write the residue information to
        res : :class:`ResidueTemplate` or :class:`ResidueTemplateContainer`
            The residue template (or template container) to write to the file
        """
        if isinstance(res, ResidueTemplate):
            # Put it into a template container with the same name
            tmp = ResidueTemplateContainer(res.name)
            tmp.append(res)
            res = tmp
        dest.write('!entry.%s.unit.atoms table  str name  str type  int typex  '
                   'int resx  int flags  int seq  int elmnt  dbl chg\n' %
                   res.name)
        for i, r in enumerate(res):
            for atom in r:
                dest.write(' "%s" "%s" 0 %d 131072 %d %d %.6f\n' % (atom.name,
                           atom.type, i+1, atom.idx+1, atom.atomic_number,
                           atom.charge))
        dest.write('!entry.%s.unit.atomspertinfo table  str pname  str ptype  '
                   'int ptypex  int pelmnt  dbl pchg\n' % res.name)
        for r in res:
            for atom in r:
                dest.write(' "%s" "%s" 0 -1 0.0\n' % (atom.name, atom.type))
        dest.write('!entry.%s.unit.boundbox array dbl\n' % res.name)
        if res.box is None:
            dest.write((' -1.000000\n' + ' 0.0\n' * 4))
        else:
            dest.write(' 1.000000\n')
            if res.box[3] == res.box[4] == res.box[5]:
                dest.write(' %f\n' % res.box[3])
            else:
                raise RuntimeError('Cannot write boxes with different angles')
            dest.write(' %f\n' % res.box[0])
            dest.write(' %f\n' % res.box[1])
            dest.write(' %f\n' % res.box[2])
        dest.write('!entry.%s.unit.childsequence single int\n %d\n' %
                   (res.name, len(res)+1))
        dest.write('!entry.%s.unit.connect array int\n' % res.name)
        if len(res) > 1:
            dest.write(' 0\n 0\n')
        else:
            if res[0].head is not None:
                dest.write(' %d\n' % (res[0].head.idx + 1))
            else:
                dest.write(' 0\n')
            if res[0].tail is not None:
                dest.write(' %d\n' % (res[0].tail.idx + 1))
            else:
                dest.write(' 0\n')
        if any(len(r) > 1 for r in res):
            dest.write('!entry.%s.unit.connectivity table  int atom1x  '
                       'int atom2x  int flags\n' % res.name)
            base = 1
            for r in res:
                for bond in r.bonds:
                    dest.write(' %d %d 1\n' % (bond.atom1.idx+base,
                                               bond.atom2.idx+base))
                base += len(r)
        dest.write('!entry.%s.unit.hierarchy table  str abovetype  int '
                   'abovex  str belowtype  int belowx\n' % res.name)
        c = 1
        for i, r in enumerate(res):
            dest.write(' "U" 0 "R" %d\n' % (i+1))
            for atom in r:
                dest.write(' "R" %d "A" %d\n' % (i+1, c))
                c += 1
        dest.write('!entry.%s.unit.name single str\n' % res.name)
        dest.write(' "%s"\n' % res.name)
        dest.write('!entry.%s.unit.positions table  dbl x  dbl y  dbl z\n' %
                   res.name)
        for r in res:
            for atom in r:
                dest.write(' %.6g %.6g %.6g\n' % (atom.xx, atom.xy, atom.xz))
        dest.write('!entry.%s.unit.residueconnect table  int c1x  int c2x  '
                   'int c3x  int c4x  int c5x  int c6x\n' % res.name)
        c = 1
        for r in res:
            # Make the CONECT1 and 0 default to first and last atom so that the
            # TREE gets set correctly by tleap. Not used for anything else...
            conn = [c, c+len(r)-1, 0, 0, 0, 0]
            if r.head is not None: conn[0] = r.head.idx + 1
            if r.tail is not None: conn[1] = r.tail.idx + 1
            for i, at in enumerate(r.connections[:4]):
                conn[i+2] = at.idx + 1
            dest.write(' %d %d %d %d %d %d\n' % tuple(conn))
            c += len(r)
        dest.write('!entry.%s.unit.residues table  str name  int seq  int '
                   'childseq  int startatomx  str restype  int imagingx\n' %
                   res.name)
        c = 1
        for i, r in enumerate(res):
            if r.type is PROTEIN:
                typ = 'p'
            elif r.type is NUCLEIC:
                typ = 'n'
            elif r.type is SOLVENT:
                typ='w'
            elif r.type is UNKNOWN:
                typ='?'
            else:
                warnings.warn('Unrecognized residue type %r' % r.type,
                              AmberWarning)
                typ = '?'
            dest.write(' "%s" %d %d %d "%s" %d\n' % (r.name, i+1, 1+len(r), c,
                       typ, _imaging_atom(r)+c))
            c += len(r)
        dest.write('!entry.%s.unit.residuesPdbSequenceNumber array int\n' %
                   res.name)
        for i, r in enumerate(res):
            if len(res) == 1:
                dest.write(' 0\n')
            else:
                dest.write(' %d\n' % (i+1))
        dest.write('!entry.%s.unit.solventcap array dbl\n' % res.name)
        dest.write(' -1.000000\n' + ' 0.0\n' * 4)
        dest.write('!entry.%s.unit.velocities table  dbl x  dbl y  dbl z\n' %
                   res.name)
        for r in res:
            for atom in r:
                try:
                    s = ' %g %g %g\n' % (atom.vx, atom.vy, atom.vz)
                except AttributeError:
                    dest.write(' 0.0 0.0 0.0\n')
                else:
                    dest.write(s)

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

# Helper routines
def _strip_enveloping_quotes(inp):
    """ Strips the quotation marks enveloping a string """
    if inp[0] == inp[-1] == '"' or inp[0] == inp[-1] == "'":
        return inp[1:-1]
    return inp

def _imaging_atom(res):
    """
    Determines the imaging atom for the residue. If all atoms are hydrogen
    except 1, it is the heavy atom. Otherwise, it is the atom *closest* to the
    COM of the residue
    """
    from ..geometry import center_of_mass
    #TODO implement the docstring
    found_heavy = False
    heavy_idx = -1
    for i, atom in enumerate(res):
        if atom.atomic_number > 1:
            heavy_idx = i
            if found_heavy: break
            found_heavy = True
    else:
        if heavy_idx != -1:
            return heavy_idx
        return 0 # No heavy atoms?? No imaging atom, then.
    coords = res.coordinates.reshape((len(res), 3))
    masses = np.zeros(len(res))
    for i, atom in enumerate(res):
        if atom.mass == 0:
            masses[i] = pt.Mass[pt.Element[atom.atomic_number]]
        else:
            masses[i] = atom.mass
    com = center_of_mass(coords, masses)
    diff = coords - com
    return np.argmin((diff * diff).sum(axis=1))



"""
This module contains all of the information for the titratable residues,
including reference energies, model compound pKas, and charge vectors for every
titratable residue treated.
"""
from math import log
from io import StringIO
from ..exceptions import AmberWarning, AmberError
import warnings

titratable_residues = ['AS4', 'GL4', 'CYS', 'TYR', 'HIP', 'LYS', 'DAP', 'DCP',
                       'DG', 'DT', 'AP', 'CP', 'G', 'U', 'HEH', 'PRN', 'TYX']

class _State(object):
    """ A protonation state """
    sort_by_resnum = True

    def __init__(self, charges, refene, refene_old=None, protcnt=None, pka_corr=None, eleccnt=None, eo_corr=None):
        self.charges = charges
        self.refene = refene
        self.refene_old = refene_old
        self.protcnt = protcnt
        self.pka_corr = pka_corr
        self.eleccnt = eleccnt
        self.eo_corr = eo_corr

class _ReferenceEnergy(object):
    """ Reference energies for various solvent models """
    LN_TO_LOG = log(10.0)
    KB = 0.00199
    TEMP = 300.0
    def __init__(self, igb1=None, igb2=None, igb5=None, igb7=None, igb8=None):
        self.pKa_is_set = False
        self.igb1 = igb1
        self.igb2 = igb2
        self.igb5 = igb5
        self.igb7 = igb7
        self.igb8 = igb8

    def solvent_energies(self, igb1=None, igb2=None, igb5=None, igb7=None, igb8=None):
        """
        Add solvent reference energies, copying the GB reference energies if
        none are explicitly given
        """
        self.solvent = _ReferenceEnergy(igb1, igb2, igb5, igb7, igb8)

    def dielc2_energies(self, igb1=None, igb2=None, igb5=None, igb7=None, igb8=None):
        """
        Add reference energies for a dielectric constant of 2.0. Since this has
        no reason to be near the original reference energies, do not use those
        in place of energies that aren't provided
        """
        self.dielc2 = _ReferenceEnergy(igb1, igb2, igb5, igb7, igb8)

    def set_pKa(self, pKa, deprotonated=False):
        """
        Adjusts the reference energies based on the pKa. If the reference energy
        is given for a deprotonated state, the pKa adjustment is subtracted from
        the reference energy. Otherwise, this state is a 'protonated' state, so
        the pKa adjustment is added to the reference energy
        """
        if self.pKa_is_set: return
        self.pKa_is_set = True
        factor = self.KB * self.LN_TO_LOG * self.TEMP * pKa

        if deprotonated:
            if self.igb1 is not None: self.igb1 -= factor
            if self.igb2 is not None: self.igb2 -= factor
            if self.igb5 is not None: self.igb5 -= factor
            if self.igb7 is not None: self.igb7 -= factor
            if self.igb8 is not None: self.igb8 -= factor
        else:
            if self.igb1 is not None: self.igb1 += factor
            if self.igb2 is not None: self.igb2 += factor
            if self.igb5 is not None: self.igb5 += factor
            if self.igb7 is not None: self.igb7 += factor
            if self.igb8 is not None: self.igb8 += factor
        # Now apply the correction to our solvent reference energies if we have
        # them (and dielectric-2 energies, if we have them)
        if hasattr(self, 'solvent'):
            self.solvent.set_pKa(pKa, deprotonated)
        if hasattr(self, 'dielc2'):
            self.dielc2.set_pKa(pKa, deprotonated)

class _LineBuffer(object):
    """ Buffer to add lines to the cpin file """

    CHARS_PER_LINE = 80

    def __init__(self, file):
        self.file = file
        self.linebuffer = ''

    def add_word(self, word):
        if len(self.linebuffer) + len(word) > self.CHARS_PER_LINE:
            self.file.write(f"{self.linebuffer}\n")
            self.linebuffer = f' {word}'
        else:
            self.linebuffer += word

    def add_words(self, words, space_delimited=False):
        """ Adds multiple words """
        extra = ''
        if space_delimited:
            extra = ' '
        for word in words:
            self.add_word(word + extra)

    def flush(self):
        """ Flushes this buffer to the file """
        if len(self.linebuffer) == 0:
            return
        self.file.write(self.linebuffer + '\n')
        self.linebuffer = ''

class TitratableResidue(object):
    """
    A residue with different protonation states defined for Amber for use in
    the Constant pH MD method implemented in sander
    """

    def __init__(self, resname, atom_list, typ, pka=None, eo=None):
        self.resname = resname
        self.atom_list = list(atom_list) # list of atom names
        self.states = []
        self.first_state = -1
        self.first_charge = -1
        self.typ = typ
        self.pKa = pka
        self.Eo = eo

    def _str_refenes(self, solvent=False, igb=2, dielc=1.0):
        """
        Converts all reference energies into a formatted string with a message
        saying if the energy is not set
        """
        ret_str = ''
        igb_str = f'igb{igb}'
        for state in self.states:
            if dielc == 2:
                if solvent:
                    refene = getattr(state.refene.dielc2.solvent, igb_str)
                else:
                    refene = getattr(state.refene.dielc2, igb_str)
            else:
                if solvent:
                    refene = getattr(state.refene.solvent, igb_str)
                else:
                    refene = getattr(state.refene, igb_str)

            if refene is None:
                ret_str += f"{'Not Set':>12s}"
            else:
                ret_str += f"{refene:12.5f}"

        return ret_str

    def __str__(self):
        delineator = '-' * (8 + 12 * len(self.states))
        if self.typ == "ph":
            ret_strs = [f"{self.resname:<4s}\tpKa = {self.pKa:5.1f}"]
        elif self.typ == "redox":
            ret_strs = [f"{self.resname:<4s}\tEo = {self.Eo:7.3f} V"]
        else:
            ret_strs = [f"{self.resname:<4s}"]
        ret_strs.append(
            f"{'ATOM':>8s}" + ''.join([f"STATE {i}".rjust(12) for i in range(len(self.states))])
        )
        for i, atom in enumerate(self.atom_list):
            ret_strs.append(f"{atom:>8s}" + "".join([f"{state.charges[i]:12.4f}" for state in self.states]))
        ret_strs.append(delineator)
        if self.typ == "ph" or self.typ == "phredox":
            ret_strs.append("Prot Cnt".rjust(8) + "".join([f"{state.protcnt:12d}" for state in self.states]))
            ret_strs.append(delineator)
            ret_strs.append("pKa Corr".rjust(8) + "".join([f"{state.pka_corr:12.4f}" for state in self.states]))
        if (self.typ == "phredox"):
            ret_strs.append(delineator)
        if (self.typ == "redox" or self.typ == "phredox"):
            ret_strs.append("Elec Cnt".rjust(8) + "".join([f"{state.eleccnt:12d}" for state in self.states]))
            ret_strs.append(delineator)
            ret_strs.append("Eo Corr".rjust(8) + "".join([f"{state.eo_corr:12.4f}" for state in self.states]))
        ret_strs.append(delineator)
        ret_strs.extend(["Reference Energies (ES = Explicit solvent, IS = Implicit solvent)", ""])
        ret_strs.append(f"{'igb=1 IS':8s}{self._str_refenes(False, 1)}")
        ret_strs.append(f"{'igb=2 IS':8s}{self._str_refenes(False, 2)}")
        ret_strs.append(f"{'igb=5 IS':8s}{self._str_refenes(False, 5)}")
        ret_strs.append(f"{'igb=7 IS':8s}{self._str_refenes(False, 7)}")
        ret_strs.append(f"{'igb=8 IS':8s}{self._str_refenes(False, 8)}")
        ret_strs.append(f"{'igb=1 ES':8s}{self._str_refenes(True, 1)}")
        ret_strs.append(f"{'igb=2 ES':8s}{self._str_refenes(True, 2)}")
        ret_strs.append(f"{'igb=5 ES':8s}{self._str_refenes(True, 5)}")
        ret_strs.append(f"{'igb=7 ES':8s}{self._str_refenes(True, 7)}")
        ret_strs.append(f"{'igb=8 ES':8s}{self._str_refenes(True, 8)}")
        ret_strs.append(delineator)
        ret_strs.extend(["Reference Energies for Internal Dielectric of 2.0", ""])
        ret_strs.append(f"{'igb=1 IS':8s}{self._str_refenes(False, 1, 2)}")
        ret_strs.append(f"{'igb=2 IS':8s}{self._str_refenes(False, 2, 2)}")
        ret_strs.append(f"{'igb=5 IS':8s}{self._str_refenes(False, 5, 2)}")
        ret_strs.append(f"{'igb=7 IS':8s}{self._str_refenes(False, 7, 2)}")
        ret_strs.append(f"{'igb=8 IS':8s}{self._str_refenes(False, 8, 2)}")
        ret_strs.append(f"{'igb=1 ES':8s}{self._str_refenes(True, 1, 2)}")
        ret_strs.append(f"{'igb=2 ES':8s}{self._str_refenes(True, 2, 2)}")
        ret_strs.append(f"{'igb=5 ES':8s}{self._str_refenes(True, 5, 2)}")
        ret_strs.append(f"{'igb=7 ES':8s}{self._str_refenes(True, 7, 2)}")
        ret_strs.append(f"{'igb=8 ES':8s}{self._str_refenes(True, 8, 2)}")
        return "\n".join(ret_strs) + "\n"

    def add_state(self, charges, refene, refene_old=None, protcnt=None, pka_corr=None, eleccnt=None, eo_corr=None):
        """ Add a single titratable state for this titratable residue """
        new_state = _State(charges, refene, refene_old=refene_old, protcnt=protcnt, pka_corr=pka_corr, eleccnt=eleccnt, eo_corr=eo_corr)
        if len(new_state.charges) != len(self.atom_list):
            raise AmberError('Wrong number of charges for new state')
        self.states.append(new_state)

    def add_states(self, charges, refenes, refenes_old=None, protcnts=None, pka_corrs=None, eleccnts=None, eo_corrs=None):
        """ Add multiple titratable states for this titratable residue """
        if len(charges) != len(refenes) or (len(charges) != len(refenes_old) and refenes_old) or (len(charges) != len(protcnts) and protcnts) \
           or (len(charges) != len(pka_corrs) and pka_corrs) or (len(charges) != len(eleccnts) and eleccnts) \
           or (len(charges) != len(eo_corrs) and eo_corrs):
            raise AmberError('Inconsistent list of parameters for TitratableResidue.add_states')
        for i in range(len(charges)):
            self.add_state(charges[i], refenes[i], refenes_old[i], protcnts[i], pka_corrs[i], eleccnts[i], eo_corrs[i])

    def cpin_pointers(self, first_atom):
        """ Sets and returns the cpin info """
        if self.first_state == -1 or self.first_charge == -1:
            raise AmberError('Must set residue pointers before writing cpin info!')
        return {'FIRST_ATOM': first_atom,
                'FIRST_CHARGE': self.first_charge,
                'FIRST_STATE': self.first_state,
                'NUM_ATOMS': len(self.atom_list),
                'NUM_STATES': len(self.states)}

    def set_first_state(self, index):
        """ Sets the first state index """
        # Has the first state already been set?
        if self.first_state != -1:
            if index != self.first_state:
                raise AmberError('First state already set differently')
        self.first_state = index

    def set_first_charge(self, index):
        """ Sets the first charge index """
        # Has it already been set?
        if self.first_charge != -1:
            if index != self.first_charge:
                raise AmberError('First charge already set differently')
        self.first_charge = index

    def reset(self):
        """ Resets the pointers """
        self.first_state = -1
        self.first_charge = -1

    def check(self):
        """ Checks that the charges are consistent w/ the protonation states """
        sum_charges = [sum(state.charges) for state in self.states]
        if (self.typ == "ph" or self.typ == "phredox"):
            protcnts = [state.protcnt for state in self.states]
        if (self.typ == "redox" or self.typ == "phredox"):
            eleccnts = [state.eleccnt for state in self.states]
        # All we have to do is make sure that the charges/proton counts are
        # consistent between the first state and every other state
        for i in range(1, len(sum_charges)):
            charge_diff = sum_charges[i] - sum_charges[0]
            if (self.typ == "ph"):
                diff = protcnts[i] - protcnts[0]
            elif (self.typ == "redox"):
                diff = eleccnts[0] - eleccnts[i]
            elif (self.typ == "phredox"):
                diff = protcnts[i] - protcnts[0] + eleccnts[0] - eleccnts[i]
            if abs(charge_diff - diff) >= 0.0001:
                warnings.warn(f'Charge definitions inconsistent in {self.resname}', AmberWarning)
        # Check all of the reference energies to make sure that the pKa was set
        # for all but one of them
        notset = 0
        valid = True
        for state in self.states:
            if not state.refene_old:
                valid = False
                break
            notset += int(state.refene_old.pKa_is_set)
        if (notset != len(self.states) - 1 and valid):
            warnings.warn(f"Not enough states are pKa-adjusted in {self.resname}")

class TitratableResidueList(list):
    """ List of all titratable residues """
    def __init__(self, system_name='Unknown', solvated=False, first_solvent=0):
        list.__init__(self)
        self.first_atoms = []
        self.residue_nums = []
        self.resstates = []
        self.system_name = system_name
        self.solvated = solvated
        self.first_sol = first_solvent

    def add_residue(self, residue, resnum, first_atom, state=0):
        """ Adds a residue to the list """
        list.append(self, residue)
        self.first_atoms.append(first_atom)
        self.residue_nums.append(resnum)
        if state < 0 or state >= len(residue.states):
            raise AmberError(
                f"Residue {residue.resname} only has states 0-{len(residue.states)} ({state} chosen)"
            )
        self.resstates.append(state)

    def set_states(self, statelist):
        """
        Sets the initial protonation states from a list -- make sure there are
        enough states in the list to set every residue, or emit a warning
        """
        if len(statelist) != len(self):
            warnings.warn(
                f"Number of states ({len(statelist)}) does not equal number of "
                f"residues ({len(self)}). Using default initial states.", AmberWarning
            )
            return
        # Check that all states are allowable
        for i, state in enumerate(statelist):
            if state < 0 or state >= len(self[i].states):
                raise AmberError(f"Bad state choice ({state}). Minimum is 0, maximum is {len(self[i].states)}")
        # If we got here, then we are OK
        self.resstates = statelist

    def sort(self):
        """ Sorts by residue number """
        atoms_residues = sorted(list(zip(self.first_atoms, self.residue_nums)))
        self.first_atoms = [atom_residue[0] for atom_residue in atoms_residues]
        self.residue_nums = [atom_residue[1] for atom_residue in atoms_residues]

    def write_cpin(self, output, igb=2, intdiel=1.0, oldfmt=False, typ="ph", coions=False):
        """ Writes the CPIN file based on the titrated residues """
        end = StringIO()
        # Reset all residues
        for res in self: res.reset()
        # Sort our residue list
        self.sort()
        limit_buf = _LineBuffer(output)
        buf = _LineBuffer(end)
        limit_buf.add_word('&CNSTPHE_LIMITS')
        limit_buf.flush()
        limit_buf.add_word(f' ntres={len(self)},')
        limit_buf.add_word(f' maxh={max(len(r.states) for r in self)},')
        if (typ == "ph"):
            buf.add_word('&CNSTPH')
        elif (typ == "redox"):
            buf.add_word('&CNSTE')
        elif (typ == "phredox"):
            buf.add_word('&CNSTPHE')
        buf.flush()
        buf.add_word(' CHRGDAT=')
        charges, energies, protcnts, pka_corrs, eleccnts, eo_corrs, pointers = [], [], [], [], [], [], []
        first_charge = 0
        first_state = 0
        for i, res in enumerate(self):
            if res.first_charge == -1:
                res.set_first_charge(first_charge)
                res.set_first_state(first_state)

                for state in res.states:
                    # See which dielectric reference energies we want
                    if intdiel == 2:
                        if (not oldfmt):
                            refene = state.refene.dielc2
                        else:
                            refene = state.refene_old.dielc2
                    else:
                        if (not oldfmt):
                            refene = state.refene
                        else:
                            refene = state.refene_old
                    # See if we want the explicit solvent refene or not
                    if self.solvated:
                        energies.append(getattr(refene.solvent, f"igb{igb}"))
                    else:
                        energies.append(getattr(refene, f"igb{igb}"))
                    if (typ == "ph" or typ == "phredox"):
                        # Add protonation count of this state
                        if (state.protcnt):
                            protcnts.append(state.protcnt)
                        else:
                            protcnts.append(0)
                        # Add pka reference of this state
                        if (state.pka_corr):
                            pka_corrs.append(state.pka_corr)
                        else:
                            pka_corrs.append(0.0)
                    if (typ == "redox" or typ == "phredox"):
                        # Add electron count of this state
                        if (state.eleccnt):
                            eleccnts.append(state.eleccnt)
                        else:
                            eleccnts.append(0)
                        # Add Eo reference of this state
                        if (state.eo_corr):
                            eo_corrs.append(state.eo_corr)
                        else:
                            eo_corrs.append(0.0)

                first_state += len(res.states)
                new_charges = []
                for state in res.states:
                    new_charges.extend(state.charges)
                charges.extend(new_charges)
                first_charge += len(new_charges)
            pointers.append(res.cpin_pointers(self.first_atoms[i]))

        limit_buf.add_word(f' natchrg={len(charges)},')
        limit_buf.add_word(f' ntstates={max(len(protcnts), len(eleccnts))},')
        limit_buf.flush()
        limit_buf.add_word('/')
        limit_buf.flush()
        # Print the charges
        for charge in charges:
            buf.add_word(f'{charge},')
        buf.flush()
        # Print the protcnts
        if (typ == "ph" or typ == "phredox"):
            buf.add_word(' PROTCNT=')
            for protcnt in protcnts:
                buf.add_word(f'{protcnt:d},')
        buf.flush()
        if (typ == "redox" or typ == "phredox"):
            buf.add_word(' ELECCNT=')
            for eleccnt in eleccnts:
                buf.add_word('%d,' % eleccnt)
        buf.flush()
        # Print the residue names
        buf.add_word(f" RESNAME='System: {self.system_name}',")
        for i, res in enumerate(self):
            buf.add_word(f"'Residue: {res.resname} {self.residue_nums[i]:d}',")
        buf.flush()
        # Print the residue states
        buf.add_word(" RESSTATE=")
        for state in self.resstates:
            buf.add_word(f'{state:d},')
        buf.flush()
        # Print the residue pointers
        buf.add_word(' ') # get a leading space
        for i, p in enumerate(pointers):
            buf.add_word(f"STATEINF({i})%FIRST_ATOM={p['FIRST_ATOM']}, ")
            buf.add_word(f"STATEINF({i})%FIRST_CHARGE={p['FIRST_CHARGE']}, ")
            buf.add_word(f"STATEINF({i})%FIRST_STATE={p['FIRST_STATE']}, ")
            buf.add_word(f"STATEINF({i})%NUM_ATOMS={p['NUM_ATOMS']}, ")
            buf.add_word(f"STATEINF({i})%NUM_STATES={p['NUM_STATES']}, ")
        buf.flush()
        # Print the reference energies
        buf.add_word(' STATENE=')
        for i, energy in enumerate(energies):
            if energy is None:
                raise AmberError(f"{i}'th reference energy not known for igb = {igb}")
            buf.add_word(f'{energy:.6f},')
        buf.flush()
        # Print the pKa or Eo reference
        if not oldfmt:
            if typ == "ph" or typ == "phredox":
                buf.add_word(' PKA_CORR=')
                for pka_corr in pka_corrs:
                    buf.add_word(f'{pka_corr:.4f},')
            buf.flush()
            if (typ == "redox" or typ == "phredox"):
                buf.add_word(' EO_CORR=')
                for eo_corr in eo_corrs:
                    buf.add_word(f'{eo_corr:.4f},')
            buf.flush()
        # Print the # of residues and explicit solvent info if required
        buf.add_word(' TRESCNT=%d,' % len(self))
        if self.solvated:
            if (typ == "ph"):
                buf.add_word(f'CPHFIRST_SOL={self.first_sol}, CPH_IGB={igb}, CPH_INTDIEL={intdiel}, ')
            elif (typ == "redox"):
                buf.add_word(f'CEFIRST_SOL={self.first_sol}, CE_IGB={igb}, CE_INTDIEL={intdiel}, ')
            elif (typ == "phredox"):
                buf.add_word(f'CPHEFIRST_SOL={self.first_sol}, CPHE_IGB={igb}, CPHE_INTDIEL={intdiel}, ')
            buf.flush()
            # Now scan through all of the waters
        buf.flush()
        buf.add_word('/')
        buf.flush()
        end.seek(0)
        output.write(end.read())

# Now define all of the titratable residues

# Aspartate
refene1 = _ReferenceEnergy(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.dielc2_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.dielc2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene2 = _ReferenceEnergy(igb1=21.4298008, igb2=26.8894581, igb5=26.5980488,
                 igb7=23.4181107, igb8=26.3448911)
refene2.solvent_energies(igb2=33.2613028, igb5=32.064349, igb7=28.262350, igb8=31.286037)
refene2.dielc2_energies(igb2=12.676908, igb5=13.084913)
refene2.dielc2.solvent_energies()
# Copying the reference energy to be printted on the old CPIN format
refene2_old = _ReferenceEnergy(igb1=21.4298008, igb2=26.8894581, igb5=26.5980488,
                 igb7=23.4181107, igb8=26.3448911)
refene2_old.solvent_energies(igb2=33.2613028, igb5=32.064349, igb7=28.262350, igb8=31.286037)
refene2_old.dielc2_energies(igb2=12.676908, igb5=13.084913)
refene2_old.dielc2.solvent_energies()
refene2_old.set_pKa(4.0, deprotonated=False)

AS4 = TitratableResidue('AS4', ['N', 'H', 'CA', 'HA', 'CB', 'HB2', 'HB3', 'CG',
                        'OD1', 'OD2', 'HD21', 'C', 'O', 'HD22', 'HD11', 'HD12'],
                        pka=4.0, typ="ph")
AS4.add_state(protcnt=0, refene=refene1, refene_old=refene1, pka_corr=0.0, # deprotonated
              charges=[-0.4157, 0.2719, 0.0341, 0.0864, -0.1783, -0.0122,
              -0.0122, 0.7994, -0.8014, -0.8014, 0.0, 0.5973, -0.5679, 0.0, 0.0,
              0.0])
AS4.add_state(protcnt=1, refene=refene2, refene_old=refene2_old, pka_corr=4.0, # protonated syn-O2
              charges=[-0.4157, 0.2719, 0.0341, 0.0864, -0.0316, 0.0488, 0.0488,
              0.6462, -0.5554, -0.6376, 0.4747, 0.5973, -0.5679, 0.0, 0.0, 0.0])
AS4.add_state(protcnt=1, refene=refene2, refene_old=refene2_old, pka_corr=4.0, # protonated anti-O2
              charges=[-0.4157, 0.2719, 0.0341, 0.0864, -0.0316, 0.0488, 0.0488,
              0.6462, -0.5554, -0.6376, 0.0, 0.5973, -0.5679, 0.4747, 0.0, 0.0])
AS4.add_state(protcnt=1, refene=refene2, refene_old=refene2_old, pka_corr=4.0, # protonated syn-O1
              charges=[-0.4157, 0.2719, 0.0341, 0.0864, -0.0316, 0.0488, 0.0488,
              0.6462, -0.6376, -0.5554, 0.0, 0.5973, -0.5679, 0.0, 0.4747, 0.0])
AS4.add_state(protcnt=1, refene=refene2, refene_old=refene2_old, pka_corr=4.0, # protonated anti-O1
              charges=[-0.4157, 0.2719, 0.0341, 0.0864, -0.0316, 0.0488, 0.0488,
              0.6462, -0.6376, -0.5554, 0.0, 0.5973, -0.5679, 0.0, 0.0, 0.4747])
AS4.check() # check that everything is consistent

# Glutamate
refene1 = _ReferenceEnergy(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.dielc2_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.dielc2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene2 = _ReferenceEnergy(igb1=3.89691326, igb2=8.4057785, igb5=8.0855764,
               igb7=5.305949, igb8=8.3591335)
refene2.solvent_energies(igb2=15.20019319, igb5=13.533409, igb7=10.682953, igb8=13.242410)
refene2.dielc2_energies(igb2=3.455596, igb5=3.957270)
refene2.dielc2.solvent_energies()
# Copying the reference energy to be printted on the old CPIN format
refene2_old = _ReferenceEnergy(igb1=3.89691326, igb2=8.4057785, igb5=8.0855764,
               igb7=5.305949, igb8=8.3591335)
refene2_old.solvent_energies(igb2=15.20019319, igb5=13.533409, igb7=10.682953, igb8=13.242410)
refene2_old.dielc2_energies(igb2=3.455596, igb5=3.957270)
refene2_old.dielc2.solvent_energies()
refene2_old.set_pKa(4.4, deprotonated=False)

GL4 = TitratableResidue('GL4', ['N', 'H', 'CA', 'HA', 'CB', 'HB2', 'HB3', 'CG',
                        'HG2', 'HG3', 'CD', 'OE1', 'OE2', 'HE21', 'C', 'O',
                        'HE22', 'HE11', 'HE12'], pka=4.4, typ="ph")
GL4.add_state(protcnt=0, refene=refene1, refene_old=refene1, pka_corr=0.0, # deprotonated
              charges=[-0.4157, 0.2719, 0.0145, 0.0779, -0.0398, -0.0173,
              -0.0173, 0.0136, -0.0425, -0.0425, 0.8054, -0.8188, -0.8188, 0.0,
              0.5973, -0.5679, 0.0, 0.0, 0.0])
GL4.add_state(protcnt=1, refene=refene2, refene_old=refene2_old, pka_corr=4.4, # protonated syn-O2
              charges=[-0.4157, 0.2719, 0.0145, 0.0779, -0.0071, 0.0256, 0.0256,
              -0.0174, 0.0430, 0.0430, 0.6801, -0.5838, -0.6511, 0.4641, 0.5973,
              -0.5679, 0.0, 0.0, 0.0])
GL4.add_state(protcnt=1, refene=refene2, refene_old=refene2_old, pka_corr=4.4, # protonated anti-O2
              charges=[-0.4157, 0.2719, 0.0145, 0.0779, -0.0071, 0.0256, 0.0256,
              -0.0174, 0.0430, 0.0430, 0.6801, -0.5838, -0.6511, 0.0, 0.5973,
              -0.5679, 0.4641, 0.0, 0.0])
GL4.add_state(protcnt=1, refene=refene2, refene_old=refene2_old, pka_corr=4.4, # protonated syn-O1
              charges=[-0.4157, 0.2719, 0.0145, 0.0779, -0.0071, 0.0256, 0.0256,
              -0.0174, 0.0430, 0.0430, 0.6801, -0.6511, -0.5838, 0.0, 0.5973,
              -0.5679, 0.0, 0.4641, 0.0])
GL4.add_state(protcnt=1, refene=refene2, refene_old=refene2_old, pka_corr=4.4, # protonated syn-O2
              charges=[-0.4157, 0.2719, 0.0145, 0.0779, -0.0071, 0.0256, 0.0256,
              -0.0174, 0.0430, 0.0430, 0.6801, -0.6511, -0.5838, 0.0, 0.5973,
              -0.5679, 0.0, 0.0, 0.4641])
GL4.check()

# Tyrosine
refene1 = _ReferenceEnergy(igb2=0, igb5=0, igb8=0)
refene1.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.dielc2_energies(igb2=0, igb5=0, igb8=0)
refene1.dielc2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene2 = _ReferenceEnergy(igb2=-65.113428, igb5=-64.166385, igb8=-61.3305355)
refene2.solvent_energies(igb2=-65.003415, igb5=-64.047229)
refene2.dielc2_energies(igb2=-32.167520, igb5=-31.751177)
refene2.dielc2.solvent_energies()
# Copying the reference energy to be printted on the old CPIN format
refene2_old = _ReferenceEnergy(igb2=-65.113428, igb5=-64.166385, igb8=-61.3305355)
refene2_old.solvent_energies(igb2=-65.003415, igb5=-64.047229)
refene2_old.dielc2_energies(igb2=-32.167520, igb5=-31.751177)
refene2_old.dielc2.solvent_energies()
refene2_old.set_pKa(9.6, deprotonated=True)

TYR = TitratableResidue('TYR', ['N', 'H', 'CA', 'HA', 'CB', 'HB2', 'HB3', 'CG',
                        'CD1', 'HD1', 'CE1', 'HE1', 'CZ', 'OH', 'HH', 'CE2',
                        'HE2', 'CD2', 'HD2', 'C', 'O'], pka=9.6, typ="ph")
TYR.add_state(protcnt=1, refene=refene1, refene_old=refene1, pka_corr=9.6, # protonated
              charges=[-0.4157, 0.2719, -0.0014, 0.0876, -0.0152, 0.0295,
              0.0295, -0.0011, -0.1906, 0.1699, -0.2341, 0.1656, 0.3226,
              -0.5579, 0.3992, -0.2341, 0.1656, -0.1906, 0.1699, 0.5973,
              -0.5679])
TYR.add_state(protcnt=0, refene=refene2, refene_old=refene2_old, pka_corr=0.0, # deprotonated
              charges=[-0.4157, 0.2719, -0.0014, 0.0876, -0.0858, 0.0190,
              0.0190, -0.2130, -0.1030, 0.1320, -0.4980, 0.1320, 0.7770,
              -0.8140, 0.0, -0.4980, 0.1320, -0.1030, 0.1320, 0.5973, -0.5679])
TYR.check()

# Histidine
refene1 = _ReferenceEnergy(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.dielc2_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.dielc2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene2 = _ReferenceEnergy(igb1=-4.208863, igb2=-2.84183, igb5=-2.86001,
               igb7=-1.741947, igb8=-3.4000)
refene2.solvent_energies(igb2=-2.77641, igb5=-2.90517)
refene2.dielc2_energies(igb2=-1.628110, igb5=-1.691093)
refene2.dielc2.solvent_energies()
# Copying the reference energy to be printted on the old CPIN format
refene2_old = _ReferenceEnergy(igb1=-4.208863, igb2=-2.84183, igb5=-2.86001,
               igb7=-1.741947, igb8=-3.4000)
refene2_old.solvent_energies(igb2=-2.77641, igb5=-2.90517)
refene2_old.dielc2_energies(igb2=-1.628110, igb5=-1.691093)
refene2_old.dielc2.solvent_energies()
refene2_old.set_pKa(6.5, deprotonated=True)
refene3 = _ReferenceEnergy(igb1=-8.230643, igb2=-6.58793, igb5=-6.70726,
               igb7=-5.118453, igb8=-6.3190)
refene3.solvent_energies(igb2=-6.483630, igb5=-6.82684)
refene3.dielc2_energies(igb2=-3.444200, igb5=-3.070113)
refene3.dielc2.solvent_energies()
# Copying the reference energy to be printted on the old CPIN format
refene3_old = _ReferenceEnergy(igb1=-8.230643, igb2=-6.58793, igb5=-6.70726,
               igb7=-5.118453, igb8=-6.3190)
refene3_old.solvent_energies(igb2=-6.483630, igb5=-6.82684)
refene3_old.dielc2_energies(igb2=-3.444200, igb5=-3.070113)
refene3_old.dielc2.solvent_energies()
refene3_old.set_pKa(7.1, deprotonated=True)

HIP = TitratableResidue('HIP', ['N', 'H', 'CA', 'HA', 'CB', 'HB2', 'HB3', 'CG',
                        'ND1', 'HD1', 'CE1', 'HE1', 'NE2', 'HE2', 'CD2', 'HD2',
                        'C', 'O'], pka=6.6, typ="ph")
HIP.add_state(protcnt=2, refene=refene1, refene_old=refene1, pka_corr=7.1, # HIP
              charges=[-0.3479, 0.2747, -0.1354, 0.1212, -0.0414, 0.0810,
              0.0810, -0.0012, -0.1513, 0.3866, -0.0170, 0.2681, -0.1718,
              0.3911, -0.1141, 0.2317, 0.7341, -0.5894])
HIP.add_state(protcnt=1, refene=refene2, refene_old=refene2_old, pka_corr=0.6, # HID
              charges=[-0.3479, 0.2747, -0.1354, 0.1212, -0.1110, 0.0402,
              0.0402, -0.0266, -0.3811, 0.3649, 0.2057, 0.1392, -0.5727, 0.0,
              0.1292, 0.1147, 0.7341, -0.5894])
HIP.add_state(protcnt=1, refene=refene3, refene_old=refene3_old, pka_corr=0.0, # HIE
              charges=[-0.3479, 0.2747, -0.1354, 0.1212, -0.1012, 0.0367,
              0.0367, 0.1868, -0.5432, 0.0, 0.1635, 0.1435, -0.2795, 0.3339,
              -0.2207, 0.1862, 0.7341, -0.5894])
HIP.check()

# Lysine
refene1 = _ReferenceEnergy(igb2=-15.2423959, igb5=-14.5392838, igb8=-18.393654)
refene1.solvent_energies(igb2=-15.1417977, igb5=-14.3152107)
refene1.dielc2_energies(igb2=-7.239587, igb5=-6.825997)
refene1.dielc2.solvent_energies()
# Copying the reference energy to be printted on the old CPIN format
refene1_old = _ReferenceEnergy(igb2=-15.2423959, igb5=-14.5392838, igb8=-18.393654)
refene1_old.solvent_energies(igb2=-15.1417977, igb5=-14.3152107)
refene1_old.dielc2_energies(igb2=-7.239587, igb5=-6.825997)
refene1_old.dielc2.solvent_energies()
refene1_old.set_pKa(10.4, deprotonated=False)
refene2 = _ReferenceEnergy(igb2=0, igb5=0, igb8=0)
refene2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene2.dielc2_energies(igb2=0, igb5=0, igb8=0)
refene2.dielc2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)

LYS = TitratableResidue('LYS', ['N', 'H', 'CA', 'HA', 'CB', 'HB2', 'HB3', 'CG',
                        'HG2', 'HG3', 'CD', 'HD2', 'HD3', 'CE', 'HE2' ,'HE3',
                        'NZ', 'HZ1', 'HZ2', 'HZ3', 'C', 'O'], pka=10.4, typ="ph")
LYS.add_state(protcnt=3, refene=refene1, refene_old=refene1_old, pka_corr=10.4, # protonated
              charges=[-0.3479, 0.2747, -0.2400, 0.1426, -0.0094, 0.0362,
              0.0362, 0.0187, 0.0103, 0.0103, -0.0479, 0.0621, 0.0621, -0.0143,
              0.1135, 0.1135, -0.3854, 0.3400, 0.3400, 0.3400, 0.7341, -0.5894])
LYS.add_state(protcnt=2, refene=refene2, refene_old=refene2, pka_corr=0.0, # deprotonated
              charges=[-0.3479, 0.2747, -0.2400, 0.1426, -0.10961, 0.0340,
              0.0340, 0.06612, 0.01041, 0.01041, -0.03768, 0.01155, 0.01155,
              0.32604, -0.03358, -0.03358, -1.03581, 0.0, 0.38604, 0.38604,
              0.7341, -0.5894])
LYS.check()

# Cysteine
refene1 = _ReferenceEnergy(igb2=77.4666763, igb5=76.2588331, igb8=71.5804519)
refene1.solvent_energies(igb2=77.6041407, igb5=76.2827217)
refene1.dielc2_energies(igb2=38.090523, igb5=37.454637)
refene1.dielc2.solvent_energies(igb2=38.489170)
# Copying the reference energy to be printted on the old CPIN format
refene1_old = _ReferenceEnergy(igb2=77.4666763, igb5=76.2588331, igb8=71.5804519)
refene1_old.solvent_energies(igb2=77.6041407, igb5=76.2827217)
refene1_old.dielc2_energies(igb2=38.090523, igb5=37.454637)
refene1_old.dielc2.solvent_energies(igb2=38.489170)
refene1_old.set_pKa(8.5, deprotonated=False)
refene2 = _ReferenceEnergy(igb2=0, igb5=0, igb8=0)
refene2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene2.dielc2_energies(igb2=0, igb5=0, igb8=0)
refene2.dielc2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)

CYS = TitratableResidue('CYS', ['N', 'H', 'CA', 'HA', 'CB', 'HB2', 'HB3', 'SG',
                        'HG', 'C', 'O'], pka=8.5, typ="ph")
CYS.add_state(protcnt=1, refene=refene1, refene_old=refene1_old, pka_corr=8.5, # protonated
              charges=[-0.4157, 0.2719, 0.0213, 0.1124, -0.1231, 0.1112, 0.1112,
                       -0.3119, 0.1933, 0.5973, -0.5679])
CYS.add_state(protcnt=0, refene=refene2, refene_old=refene2, pka_corr=0.0, # deprotonated
              charges=[-0.4157, 0.2719, 0.0213, 0.1124, -0.3593, 0.1122, 0.1122,
                       -0.8844, 0.0, 0.5973, -0.5679])
CYS.check()

# Deoxy-adenine
refene1 = _ReferenceEnergy(igb2=-19.8442, igb5=-19.8442)
refene1.solvent_energies()
refene1.dielc2_energies(igb2=-9.106013, igb5=-9.404867)
refene1.dielc2.solvent_energies(igb2=-9.779586)
# Copying the reference energy to be printted on the old CPIN format
refene1_old = _ReferenceEnergy(igb2=-19.8442, igb5=-19.8442)
refene1_old.solvent_energies()
refene1_old.dielc2_energies(igb2=-9.106013, igb5=-9.404867)
refene1_old.dielc2.solvent_energies(igb2=-9.779586)
refene1_old.set_pKa(3.9, deprotonated=True)
refene2 = _ReferenceEnergy(igb2=0, igb5=0, igb8=0)
refene2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene2.dielc2_energies(igb2=0, igb5=0, igb8=0)
refene2.dielc2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)

DAP = TitratableResidue('DAP', ['P', 'O1P', 'O2P', "O5'", "C5'", "H5'1", "H5'2",
                        "C4'", "H4'", "O4'", "C1'", "H1'", 'N9', 'C8', 'H8',
                        'N7', 'C5', 'C6', 'N6', 'H61', 'H62', 'N1', 'C2', 'H2',
                        'N3', 'C4', "C3'", "H3'", "C2'", "H2'1", "H2'2", "O3'",
                        'H1'], pka=3.9, typ="ph")
DAP.add_state(protcnt=1, refene=refene1, refene_old=refene1_old, pka_corr=0.0, # deprotonated
              charges=[1.1659, -0.7761, -0.7761, -0.4954, -0.0069, 0.0754,
              0.0754, 0.1629, 0.1176, -0.3691, 0.0431, 0.1838, -0.0268, 0.1607,
              0.1877, -0.6175, 0.0725, 0.6897, -0.9123, 0.4167, 0.4167, -0.7624,
              0.5716, 0.0598, -0.7417, 0.38, 0.0713, 0.0985, -0.0854, 0.0718,
              0.0718, -0.5232, 0.0])
DAP.add_state(protcnt=2, refene=refene2, refene_old=refene2, pka_corr=3.9, # protonated
              charges=[1.1659, -0.7761, -0.7761, -0.4954, -0.0069, 0.0754,
              0.0754, 0.1629, 0.1176, -0.3691, 0.0431, 0.1838, 0.0944, 0.1617,
              0.2281, -0.5674, 0.1358, 0.5711, -0.8251, 0.4456, 0.4456, -0.575,
              0.4251, 0.1437, -0.5611, 0.3421, 0.0713, 0.0985, -0.0854, 0.0718,
              0.0718, -0.5232, 0.4301])
DAP.check()

# Deoxy-cytosine
refene1 = _ReferenceEnergy(igb2=-40.526, igb5=-40.526)
refene1.solvent_energies()
refene1.dielc2_energies(igb2=-19.447553, igb5=-19.842087)
refene1.dielc2.solvent_energies(igb2=-20.121129)
# Copying the reference energy to be printted on the old CPIN format
refene1_old = _ReferenceEnergy(igb2=-40.526, igb5=-40.526)
refene1_old.solvent_energies()
refene1_old.dielc2_energies(igb2=-19.447553, igb5=-19.842087)
refene1_old.dielc2.solvent_energies(igb2=-20.121129)
refene1_old.set_pKa(4.3, deprotonated=True)
refene2 = _ReferenceEnergy(igb2=0, igb5=0)
refene2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene2.dielc2_energies(igb2=0, igb5=0, igb8=0)
refene2.dielc2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)

DCP = TitratableResidue('DCP', ['P', 'O1P', 'O2P', "O5'", "C5'", "H5'1", "H5'2",
                        "C4'", "H4'", "O4'", "C1'", "H1'", 'N1', 'C6', 'H6',
                        'C5', 'H5', 'C4', 'N4', 'H41', 'H42', 'N3', 'C2', 'O2',
                        "C3'", "H3'", "C2'", "H2'1", "H2'2", "O3'", 'H3'],
                        pka=4.3, typ="ph")
DCP.add_state(protcnt=1, refene=refene1, refene_old=refene1_old, pka_corr=0.0, # deprotonated
              charges=[1.1659, -0.7761, -0.7761, -0.4954, -0.0069, 0.0754,
              0.0754, 0.1629, 0.1176, -0.3691, -0.0116, 0.1963, -0.0339,
              -0.0183, 0.2293, -0.5222, 0.1863, 0.8439, -0.9773, 0.4314, 0.4314,
              -0.7748, 0.7959, -0.6548, 0.0713, 0.0985, -0.0854, 0.0718, 0.0718,
              -0.5232, 0.0])
DCP.add_state(protcnt=2, refene=refene2, refene_old=refene2, pka_corr=4.3, # protonated
              charges=[1.1659, -0.7761, -0.7761, -0.4954, -0.0069, 0.0754,
              0.0754, 0.1629, 0.1176, -0.3691, -0.0116, 0.1963, 0.2167, -0.0282,
              0.2713, -0.4162, 0.2179, 0.6653, -0.859, 0.4598, 0.4598, -0.4956,
              0.5371, -0.5028, 0.0713, 0.0985, -0.0854, 0.0718, 0.0718, -0.5232,
              0.4108])
DCP.check()

# Deoxy-guanine
refene1 = _ReferenceEnergy(igb2=0, igb5=0, igb8=0)
refene1.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.dielc2_energies(igb2=0, igb5=0, igb8=0)
refene1.dielc2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene2 = _ReferenceEnergy(igb2=-90.0011, igb5=-90.0011)
refene2.solvent_energies()
refene2.dielc2_energies(igb2=-44.031593, igb5=-43.588343)
refene2.dielc2.solvent_energies(igb2=-45.090067)
# Copying the reference energy to be printted on the old CPIN format
refene2_old = _ReferenceEnergy(igb2=-90.0011, igb5=-90.0011)
refene2_old.solvent_energies()
refene2_old.dielc2_energies(igb2=-44.031593, igb5=-43.588343)
refene2_old.dielc2.solvent_energies(igb2=-45.090067)
refene2_old.set_pKa(9.2, deprotonated=True)

DG = TitratableResidue('DG', ['P', 'O1P', 'O2P', "O5'", "C5'", "H5'1", "H5'2",
                       "C4'", "H4'", "O4'", "C1'", "H1'", 'N9', 'C8', 'H8',
                       'N7', 'C5', 'C6', 'O6', 'N1', 'H1', 'C2', 'N2', 'H21',
                       'H22', 'N3', 'C4', "C3'", "H3'", "C2'", "H2'1", "H2'2",
                       "O3'"], pka=9.2, typ="ph")
DG.add_state(protcnt=1, refene=refene1, refene_old=refene1, pka_corr=9.2, # protonated
             charges=[1.1659, -0.7761, -0.7761, -0.4954, -0.0069, 0.0754,
             0.0754, 0.1629, 0.1176, -0.3691, 0.0358, 0.1746, 0.0577, 0.0736,
             0.1997, -0.5725, 0.1991, 0.4918, -0.5699, -0.5053, 0.352, 0.7432,
             -0.923, 0.4235, 0.4235, -0.6636, 0.1814, 0.0713, 0.0985, -0.0854,
             0.0718, 0.0718, -0.5232])
DG.add_state(protcnt=0, refene=refene2, refene_old=refene2_old, pka_corr=0.0, # deprotonated
             charges=[1.1659, -0.7761, -0.7761, -0.4954, -0.0069, 0.0754,
             0.0754, 0.1629, 0.1176, -0.3691, 0.0358, 0.1746, -0.0507, 0.0779,
             0.1516, -0.6122, 0.0806, 0.7105, -0.7253, -0.8527, 0.0, 0.9561,
             -0.9903, 0.3837, 0.3837, -0.8545, 0.2528, 0.0713, 0.0985, -0.0854,
             0.0718, 0.0718, -0.5232])
DG.check()

# Deoxy-thymine
refene1 = _ReferenceEnergy(igb2=0, igb5=0)
refene1.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.dielc2_energies(igb2=0, igb5=0, igb8=0)
refene1.dielc2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene2 = _ReferenceEnergy(igb2=-56.7729, igb5=-56.7729)
refene2.solvent_energies(igb2=-28.429391)
refene2.dielc2_energies(igb2=-28.085730, igb5=-27.298290)
refene2.dielc2.solvent_energies()
# Copying the reference energy to be printted on the old CPIN format
refene2_old = _ReferenceEnergy(igb2=-56.7729, igb5=-56.7729)
refene2_old.solvent_energies(igb2=-28.429391)
refene2_old.dielc2_energies(igb2=-28.085730, igb5=-27.298290)
refene2_old.dielc2.solvent_energies()
refene2_old.set_pKa(9.7, deprotonated=True)

DT = TitratableResidue('DT', ['P', 'O1P', 'O2P', "O5'", "C5'", "H5'1", "H5'2",
                       "C4'", "H4'", "O4'", "C1'", "H1'", 'N1', 'C6', 'H6',
                       'C5', 'C7', 'H71', 'H72', 'H73', 'C4', 'O4', 'N3', 'H3',
                       'C2', 'O2', "C3'", "H3'", "C2'", "H2'1", "H2'2", "O3'"],
                       pka=9.7, typ="ph")
DT.add_state(protcnt=1, refene=refene1, refene_old=refene1, pka_corr=9.7, # protonated
             charges=[1.1659, -0.7761, -0.7761, -0.4954, -0.0069, 0.0754,
             0.0754, 0.1629, 0.1176, -0.3691, 0.068, 0.1804, -0.0239, -0.2209,
             0.2607, 0.0025, -0.2269, 0.077, 0.077, 0.077, 0.5194, -0.5563,
             -0.434, 0.342, 0.5677, -0.5881, 0.0713, 0.0985, -0.0854, 0.0718,
             0.0718, -0.5232])
DT.add_state(protcnt=0, refene=refene2, refene_old=refene2_old, pka_corr=0.0, # deprotonated
             charges=[1.1659, -0.7761, -0.7761, -0.4954, -0.0069, 0.0754,
             0.0754, 0.1629, 0.1176, -0.3691, 0.068, 0.1804, -0.2861, -0.1874,
             0.2251, -0.1092, -0.2602, 0.0589, 0.0589, 0.0589, 0.8263, -0.7396,
             -0.9169, 0.0, 0.9167, -0.7722, 0.0713, 0.0985, -0.0854, 0.0718,
             0.0718, -0.5232])
DT.check()

# Adenine
refene1 = _ReferenceEnergy(igb2=0, igb5=0)
refene1.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.dielc2_energies(igb2=0, igb5=0, igb8=0)
refene1.dielc2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene2 = _ReferenceEnergy(igb2=14.851840012, igb5=15.1166001033)
refene2.solvent_energies(igb2=15.026098360790293,igb5=15.143651997474915)
refene2.solvent_energies()
refene2.dielc2_energies(igb2=6.953887, igb5=7.092043)
refene2.dielc2.solvent_energies(igb2=7.544988)
# Copying the reference energy to be printted on the old CPIN format
refene2_old = _ReferenceEnergy(igb2=14.851840012, igb5=15.1166001033)
refene2_old.solvent_energies(igb2=15.026098360790293,igb5=15.143651997474915)
refene2_old.solvent_energies()
refene2_old.dielc2_energies(igb2=6.953887, igb5=7.092043)
refene2_old.dielc2.solvent_energies(igb2=7.544988)
refene2_old.set_pKa(3.5, deprotonated=False)

AP = TitratableResidue('AP', ['P', 'O1P', 'O2P', "O5'", "C5'", "H5'1", "H5'2",
                       "C4'", "H4'", "O4'", "C1'", "H1'", 'N9', 'C8', 'H8',
                       'N7', 'C5', 'C6', 'N6', 'H61', 'H62', 'N1', 'C2', 'H2',
                       'N3', 'C4', "C3'", "H3'", "C2'", "H2'1", "O2'", "HO'2",
                       "O3'", 'H1'], pka=3.9, typ="ph")
AP.add_state(protcnt=0, refene=refene1, refene_old=refene1, pka_corr=0.0, # deprotonated
             charges=[1.1662, -0.776, -0.776, -0.4989, 0.0558, 0.0679, 0.0679,
             0.1065, 0.1174, -0.3548, 0.0394, 0.2007, -0.0251, 0.2006, 0.1553,
             -0.6073, 0.0515, 0.7009, -0.9019, 0.4115, 0.4115, -0.7615, 0.5875,
             0.0473, -0.6997, 0.3053, 0.2022, 0.0615, 0.067, 0.0972, -0.6139,
             0.4186, -0.5246, 0.0])
AP.add_state(protcnt=1, refene=refene2, refene_old=refene2_old, pka_corr=3.5, # protonated
             charges=[1.1662, -0.776, -0.776, -0.4989, 0.0558, 0.0679, 0.0679,
             0.1065, 0.1174, -0.3548, 0.0394, 0.2007, 0.0961, 0.2011, 0.1965,
             -0.5569, 0.1136, 0.5845, -0.8152, 0.4403, 0.4403, -0.5776, 0.4435,
             0.1307, -0.5201, 0.2681, 0.2022, 0.0615, 0.067, 0.0972, -0.6139,
             0.4186, -0.5246, 0.431])
AP.check()

# Cytosine
refene1 = _ReferenceEnergy(igb2=0, igb5=0)
refene1.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.dielc2_energies(igb2=0, igb5=0, igb8=0)
refene1.dielc2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene2 = _ReferenceEnergy(igb2=37.501800178, igb5=38.0081251132)
refene2.solvent_energies(igb2=37.378544257354164, igb5=37.90444570773976)
refene2.dielc2_energies(igb2=18.483513, igb5=19.016390)
refene2.dielc2.solvent_energies()
# Copying the reference energy to be printted on the old CPIN format
refene2_old = _ReferenceEnergy(igb2=37.501800178, igb5=38.0081251132)
refene2_old.solvent_energies(igb2=37.378544257354164, igb5=37.90444570773976)
refene2_old.dielc2_energies(igb2=18.483513, igb5=19.016390)
refene2_old.dielc2.solvent_energies()
refene2_old.set_pKa(4.2, deprotonated=False)

CP = TitratableResidue('CP', ['P', 'O1P', 'O2P', "O5'", "C5'", "H5'1", "H5'2",
                       "C4'", "H4'", "O4'", "C1'", "H1'", 'N1', 'C6', 'H6',
                       'C5', 'H5', 'C4', 'N4', 'H41', 'H42', 'N3', 'C2', 'O2',
                       "C3'", "H3'", "C2'", "H2'1", "O2'", "HO'2", "O3'", 'H3'],
                       pka=4.3, typ="ph")
CP.add_state(protcnt=1, refene=refene1, refene_old=refene1, pka_corr=0.0, # deprotonated
             charges=[1.1662, -0.776, -0.776, -0.4989, 0.0558, 0.0679, 0.0679,
             0.1065, 0.1174, -0.3548, 0.0066, 0.2029, -0.0484, 0.0053, 0.1958,
             -0.5215, 0.1928, 0.8185, -0.953, 0.4234, 0.4234, -0.7584, 0.7538,
             -0.6252, 0.2022, 0.0615, 0.067, 0.0972, -0.6139, 0.4186, -0.5246,
             0.0])
CP.add_state(protcnt=2, refene=refene2, refene_old=refene2_old, pka_corr=4.2, # protonated
             charges=[1.1662, -0.776, -0.776, -0.4989, 0.0558, 0.0679, 0.0679,
             0.1065, 0.1174, -0.3548, 0.0066, 0.2029, 0.1954, 0.0028, 0.2366,
             -0.4218, 0.2253, 0.6466, -0.8363, 0.4518, 0.4518, -0.4871, 0.5039,
             -0.4753, 0.2022, 0.0615, 0.067, 0.0972, -0.6139, 0.4186, -0.5246,
             0.4128])
CP.check()

# Guanine
refene1 = _ReferenceEnergy(igb2=0, igb5=0)
refene1.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.dielc2_energies(igb2=0, igb5=0, igb8=0)
refene1.dielc2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene2 = _ReferenceEnergy(igb2=-97.094725165, igb5=-96.0365352027)
refene2.solvent_energies(igb2=-97.31657849010276, igb5=-95.95654436492156)
refene2.dielc2_energies(igb2=-47.410980, igb5=-47.008233)
refene2.dielc2.solvent_energies(igb2=-48.222021)
# Copying the reference energy to be printted on the old CPIN format
refene2_old = _ReferenceEnergy(igb2=-97.094725165, igb5=-96.0365352027)
refene2_old.solvent_energies(igb2=-97.31657849010276, igb5=-95.95654436492156)
refene2_old.dielc2_energies(igb2=-47.410980, igb5=-47.008233)
refene2_old.dielc2.solvent_energies(igb2=-48.222021)
refene2_old.set_pKa(9.2, deprotonated=True)

G = TitratableResidue('G', ['P', 'O1P', 'O2P', "O5'", "C5'", "H5'1", "H5'2",
                      "C4'", "H4'", "O4'", "C1'", "H1'", 'N9', 'C8', 'H8', 'N7',
                      'C5', 'C6', 'O6', 'N1', 'H1', 'C2', 'N2', 'H21', 'H22',
                      'N3', 'C4', "C3'", "H3'", "C2'", "H2'1", "O2'", "HO'2",
                      "O3'"], pka=9.2, typ="ph")
G.add_state(protcnt=1, refene=refene1, refene_old=refene1, pka_corr=9.2, # protonated
            charges=[1.1662, -0.776, -0.776, -0.4989, 0.0558, 0.0679, 0.0679,
            0.1065, 0.1174, -0.3548, 0.0191, 0.2006, 0.0492, 0.1374, 0.164,
            -0.5709, 0.1744, 0.477, -0.5597, -0.4787, 0.3424, 0.7657, -0.9672,
            0.4364, 0.4364, -0.6323, 0.1222, 0.2022, 0.0615, 0.067, 0.0972,
            -0.6139, 0.4186, -0.5246])
G.add_state(protcnt=0, refene=refene2, refene_old=refene2_old, pka_corr=0.0, # deprotonated
            charges=[1.1662, -0.776, -0.776, -0.4989, 0.0558, 0.0679, 0.0679,
            0.1065, 0.1174, -0.3548, 0.0191, 0.2006, -0.0623, 0.1479, 0.1137,
            -0.6127, 0.0488, 0.7137, -0.7191, -0.8557, 0.0, 0.9976, -1.0387,
            0.3969, 0.3969, -0.8299, 0.1992, 0.2022, 0.0615, 0.067, 0.0972,
            -0.6139, 0.4186, -0.5246])
G.check()

# Uracil
refene1 = _ReferenceEnergy(igb2=0, igb5=0)
refene1.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.dielc2_energies(igb2=0, igb5=0, igb8=0)
refene1.dielc2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene2 = _ReferenceEnergy(igb2=-136.326020191, igb5=-134.938275039)
refene2.solvent_energies(igb2=-136.5653533428478, igb5=-135.06973320905044)
refene2.dielc2_energies(igb2=-67.270690, igb5=-66.605330)
refene2.dielc2.solvent_energies()
# Copying the reference energy to be printted on the old CPIN format
refene2_old = _ReferenceEnergy(igb2=-136.326020191, igb5=-134.938275039)
refene2_old.solvent_energies(igb2=-136.5653533428478, igb5=-135.06973320905044)
refene2_old.dielc2_energies(igb2=-67.270690, igb5=-66.605330)
refene2_old.dielc2.solvent_energies()
refene2_old.set_pKa(9.2, deprotonated=True)

U = TitratableResidue('U', ['P', 'O1P', 'O2P', "O5'", "C5'", "H5'1", "H5'2",
                      "C4'", "H4'", "O4'", "C1'", "H1'", 'N1', 'C6', 'H6', 'C5',
                      'H5', 'C4', 'O4', 'N3', 'H3', 'C2', 'O2', "C3'", "H3'",
                      "C2'", "H2'1", "O2'", "HO'2", "O3'"], pka=9.3, typ="ph")
U.add_state(protcnt=1, refene=refene1, refene_old=refene1, pka_corr=9.2, # protonated
            charges=[1.1662, -0.776, -0.776, -0.4989, 0.0558, 0.0679, 0.0679,
            0.1065, 0.1174, -0.3548, 0.0674, 0.1824, 0.0418, -0.1126, 0.2188,
            -0.3635, 0.1811, 0.5952, -0.5761, -0.3549, 0.3154, 0.4687, -0.5477,
            0.2022, 0.0615, 0.067, 0.0972, -0.6139, 0.4186, -0.5246])
U.add_state(protcnt=0, refene=refene2, refene_old=refene2_old, pka_corr=0.0, # deprotonated
            charges=[1.1662, -0.776, -0.776, -0.4989, 0.0558, 0.0679, 0.0679,
            0.1065, 0.1174, -0.3548, 0.0674, 0.1824, -0.2733, 0.0264, 0.1501,
            -0.582, 0.156, 0.9762, -0.7808, -0.9327, 0.0, 0.8698, -0.7435,
            0.2022, 0.0615, 0.067, 0.0972, -0.6139, 0.4186, -0.5246])
U.check()

# HEH: HEME ring + parts of 2 HIS + 2 CYS
refene1 = _ReferenceEnergy(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.dielc2_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.dielc2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene2 = _ReferenceEnergy(igb2=-15.493731, igb5=-16.349152, igb7=-16.509509, igb8=-22.025653) # Implicit
refene2.solvent_energies(igb2=-15.209270, igb5=-15.840853, igb7=-15.495868) # Explicit
refene2.dielc2_energies()
refene2.dielc2.solvent_energies()

HEH = TitratableResidue('HEH', ['FE', 'NA', 'C1A', 'C2A', 'C3A', 'CMA', 'HMA1', 'HMA2', 'HMA3', 'C4A',
                                'CHB', 'HHB', 'C1B', 'NB', 'C2B', 'CMB', 'HMB1', 'HMB2', 'HMB3', 'C3B',
                                'CAB', 'HAB', 'CBB', 'HBB1', 'HBB2', 'HBB3', 'C4B', 'CHC', 'HHC', 'C1C',
                                'NC', 'C2C', 'CMC', 'HMC1', 'HMC2', 'HMC3', 'C3C', 'CAC', 'HAC', 'CBC',
                                'HBC1', 'HBC2', 'HBC3', 'C4C', 'CHD', 'HHD', 'C1D', 'ND', 'C2D', 'CMD',
                                'HMD1', 'HMD2', 'HMD3', 'C3D', 'C4D', 'CHA', 'HHA', 'CBC1', 'HB2C',
                                'HB3C', 'SGC1', 'CB1', 'HB21', 'HB31', 'CG1', 'ND11', 'HD11', 'CE11',
                                'HE11', 'NE21', 'CD21', 'HD21', 'CBB2', 'HB2B', 'HB3B', 'SGB2', 'CB2',
                                'HB22', 'HB32', 'CG2', 'ND12', 'HD12', 'CE12', 'HE12', 'NE22', 'CD22',
                                'HD22'],
                        eo=-0.203, typ="redox")
HEH.add_state(eleccnt=2, refene=refene1, eo_corr=0.0, # FE3+ (oxidized state)
              charges=[ 0.6660, -0.1530, -0.0956,  0.1274,  0.1624, -0.2600,  0.0743,  0.0743,  0.0743,
                       -0.0766, -0.0586,  0.1300, -0.0206, -0.2560,  0.1394, -0.2240,  0.0663,  0.0663,
                        0.0663,  0.0734, -0.0935,  0.1765, -0.4035,  0.1375,  0.1375,  0.1375,  0.0504,
                       -0.1806,  0.1430,  0.0364, -0.2390,  0.1784, -0.2325,  0.0635,  0.0635,  0.0635,
                        0.0084, -0.0570,  0.2130, -0.4090,  0.1357,  0.1357,  0.1357, -0.0266, -0.0576,
                        0.1360, -0.1156, -0.1130,  0.1604, -0.2575,  0.0752,  0.0752,  0.0752,  0.1444,
                       -0.1126,  0.0104,  0.1090, -0.3349,  0.1297,  0.1297, -0.1760, -0.0803,  0.0269,
                        0.0269,  0.1990, -0.2930,  0.3650,  0.0120,  0.1180, -0.0400, -0.2020,  0.1790,
                       -0.3349,  0.1297,  0.1297, -0.1760, -0.0803,  0.0269,  0.0269,  0.1990, -0.2930,
                        0.3650,  0.0120,  0.1180, -0.0400, -0.2020,  0.1790]
              )
HEH.add_state(eleccnt=3, refene=refene2, eo_corr=-0.203, # FE2+ (reduced state)
              charges=[ 0.4800, -0.1337, -0.1455,  0.1285,  0.1325, -0.2545,  0.0608,  0.0608,  0.0608,
                       -0.0865, -0.0815,  0.1220, -0.0425, -0.2490,  0.1605, -0.1935,  0.0422,  0.0422,
                        0.0422, -0.0025, -0.0390,  0.1720, -0.4255,  0.1348,  0.1348,  0.1348,  0.0405,
                       -0.1725,  0.1320, -0.0525, -0.1980,  0.2125, -0.1985,  0.0405,  0.0405,  0.0405,
                       -0.0355, -0.0195,  0.1915, -0.4340,  0.1320,  0.1320,  0.1320, -0.0275, -0.0805,
                        0.1290, -0.1275, -0.1020,  0.1335, -0.2525,  0.0615,  0.0615,  0.0615,  0.1415,
                       -0.1575,  0.0275,  0.0950, -0.3343,  0.1300,  0.1300, -0.2315, -0.0967,  0.0187,
                        0.0187,  0.2030, -0.3450,  0.3550,  0.0110,  0.1090, -0.0270, -0.2170,  0.1750,
                       -0.3343,  0.1300,  0.1300, -0.2315, -0.0967,  0.0187,  0.0187,  0.2030, -0.3450,
                        0.3550,  0.0110,  0.1090, -0.0270, -0.2170,  0.1750]
              )
HEH.check()

# Propionate
refene1 = _ReferenceEnergy(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.dielc2_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.dielc2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene2 = _ReferenceEnergy(igb2=10.356928, igb5=9.943308, igb7=7.020632, igb8=5.259028) # Implicit
refene2.solvent_energies(igb2=16.751825, igb5=15.661934, igb7=12.906876, igb8=15.024553) # Explicit
refene2.dielc2_energies()
refene2.dielc2.solvent_energies()
# Copying the reference energy to be printted on the old CPIN format
refene2_old = _ReferenceEnergy(igb2=10.356928, igb5=9.943308, igb7=7.020632, igb8=5.259028) # Implicit
refene2_old.solvent_energies(igb2=16.751825, igb5=15.661934, igb7=12.906876, igb8=15.024553) # Explicit
refene2_old.dielc2_energies()
refene2_old.dielc2.solvent_energies()
refene2_old.set_pKa(4.85, deprotonated=False)

PRN = TitratableResidue('PRN',
                        ['CA', 'HA1', 'HA2', 'CB', 'HB1', 'HB2', 'CG',
                        'O1', 'O2', 'H11', 'H12', 'H21', 'H22'], pka=4.85, typ="ph")

PRN.add_state(protcnt=0, refene=refene1, refene_old=refene1, pka_corr=0.0, # deprotonated
              charges=[-0.0508, -0.0173,
              -0.0173, 0.0026, -0.0425, -0.0425, 0.8054, -0.8188, -0.8188, 0.0,
              0.0, 0.0, 0.0])

PRN.add_state(protcnt=1, refene=refene2, refene_old=refene2_old, pka_corr=4.85, # protonated syn-O1
              charges=[-0.0181, 0.0256, 0.0256,
              -0.0284, 0.0430, 0.0430, 0.6801, -0.6511, -0.5838, 0.4641,
              0.0, 0.0, 0.0])

PRN.add_state(protcnt=1, refene=refene2, refene_old=refene2_old, pka_corr=4.85, # protonated anti-O1
              charges=[-0.0181, 0.0256, 0.0256,
              -0.0284, 0.0430, 0.0430, 0.6801, -0.6511, -0.5838, 0.0,
              0.4641, 0.0, 0.0])

PRN.add_state(protcnt=1, refene=refene2, refene_old=refene2_old, pka_corr=4.85, # protonated syn-O2
              charges=[-0.0181, 0.0256, 0.0256,
              -0.0284, 0.0430, 0.0430, 0.6801, -0.5838, -0.6511, 0.0,
              0.0, 0.4641, 0.0])

PRN.add_state(protcnt=1, refene=refene2, refene_old=refene2_old, pka_corr=4.85, # protonated anti-O2
              charges=[-0.0181, 0.0256, 0.0256,
              -0.0284, 0.0430, 0.0430, 0.6801, -0.5838, -0.6511, 0.0,
              0.0, 0.0, 0.4641])

PRN.check()

# Tyrosine, pH and redox active
refene1 = _ReferenceEnergy(igb2=0, igb5=0, igb8=0)
refene1.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene1.dielc2_energies(igb2=0, igb5=0, igb8=0)
refene1.dielc2.solvent_energies(igb1=0, igb2=0, igb5=0, igb7=0, igb8=0)
refene2 = _ReferenceEnergy(igb2=5.409084) # Implicit
refene2.solvent_energies(igb2=5.210075) # Explicit
refene2.dielc2_energies()
refene2.dielc2.solvent_energies()
refene3 = _ReferenceEnergy(igb2=-52.293859) # Implicit
refene3.solvent_energies(igb2=-52.185005) # Explicit
refene3.dielc2_energies()
refene3.dielc2.solvent_energies()
refene4 = _ReferenceEnergy(igb2=24.166408) # Implicit
refene4.solvent_energies(igb2=24.224913) # Explicit
refene4.dielc2_energies()
refene4.dielc2.solvent_energies()

TYX = TitratableResidue('TYX', ['N', 'H', 'CA', 'HA', 'CB', 'HB2', 'HB3', 'CG',
                        'CD1', 'HD1', 'CE1', 'HE1', 'CZ', 'OH', 'HH', 'CE2',
                        'HE2', 'CD2', 'HD2', 'C', 'O'], typ="phredox")
# Note: pka_corr differences only make sense between states with the same number of electrons
# Note: eo_corr differences only make sense between states with the same number of protons
TYX.add_state(protcnt=1, eleccnt=1, refene=refene1, pka_corr=9.6, eo_corr=1.4, # tyrOH
              charges=[-0.4157, 0.2719, -0.0014, 0.0876, -0.1163, 0.0548,
                       0.0548, -0.0139, -0.1142, 0.1615, -0.3410, 0.1911,
                       0.4198, -0.5278, 0.3621, -0.3410, 0.1911, -0.1142,
                       0.1615, 0.5973, -0.5679])
TYX.add_state(protcnt=1, eleccnt=0, refene=refene2, pka_corr=-2.0, eo_corr=0.0, # tyrOH+
              charges=[-0.4157, 0.2719, -0.0014, 0.0876, -0.2911, 0.1123,
                       0.1123, 0.4479, -0.1923, 0.2177, -0.1736, 0.2109,
                       0.5560, -0.4376, 0.4031, -0.1736, 0.2109, -0.1923,
                       0.2177, 0.5973, -0.5679])
TYX.add_state(protcnt=0, eleccnt=1, refene=refene3, pka_corr=0.0, eo_corr=0.71, # tyrO-
              charges=[-0.4157, 0.2719, -0.0014, 0.0876, -0.0500, 0.0300,
                       0.0300, -0.1786, -0.1545, 0.1418, -0.4793, 0.1318,
                       0.7197, -0.8026, 0.0000, -0.4793, 0.1318, -0.1545,
                       0.1418, 0.5974, -0.5678])
TYX.add_state(protcnt=0, eleccnt=0, refene=refene4, pka_corr=0.0, eo_corr=0.0, # tyrO
              charges=[-0.4157, 0.2719, -0.0014, 0.0876, -0.1941, 0.0931,
                       0.0931, 0.0538, -0.1216, 0.1629, -0.3224, 0.1643,
                       0.6679, -0.4519, 0.0000, -0.3224, 0.1643, -0.1216,
                       0.1629, 0.5973, -0.5679])
TYX.check()
